r"""Bounding box oriented in world space which defines a normalized domain."""

from __future__ import annotations

from copy import copy as shallow_copy
from typing import Any, Optional, Sequence, Union, overload

import numpy as np

import torch
from torch import Tensor

from .grid import ALIGN_CORNERS, Axes, Grid
from .linalg import hmm, homogeneous_matrix, homogeneous_transform
from .tensor import as_tensor, cat_scalars
from .typing import Array, Device, Shape, Size


class Cube(object):
    r"""Bounding box oriented in world space which defines a normalized domain.

    Coordinates of points within this domain can be either with respect to the world coordinate
    system or the cube defined by the bounding box where coordinate axes are parallel to the
    cube edges and have a uniform side length of 2. The latter are the normalized coordinates
    used by ``torch.nn.functional.grid_sample()``, in particular. In terms of the coordinate
    transformations, a :class:`.Cube` is thus equivalent to a :class:`.Grid` with three points
    along each dimension and ``align_corners=True``.

    A regular sampling :class:`.Grid`, on the other hand, subsamples the world space within the bounds
    defined by the cube into a number of equally sized cells or equally spaced points, respectivey.
    How the grid points relate to the faces of the cube depends on :meth:`.Grid.align_corners`.

    """

    __slots__ = ("_center", "_direction", "_extent")

    def __init__(
        self,
        extent: Optional[Union[Array, float]],
        center: Optional[Union[Array, float]] = None,
        origin: Optional[Union[Array, float]] = None,
        direction: Optional[Array] = None,
        device: Optional[Device] = None,
    ):
        r"""Initialize cube attributes.

        Args:
            extent: Extent ``(extent_x, ...)`` of the cube in world units.
            center: Cube center point ``(x, ...)`` in world space.
            origin: World coordinates ``(x, ...)`` of lower left corner.
            direction: Direction cosines defining orientation of cube in world space.
                The direction cosines are vectors that point along the cube edges.
                Each column of the matrix indicates the direction cosines of the unit vector
                that is parallel to the cube edge corresponding to that dimension.
            device: Device on which to store attributes. Uses ``"cpu"`` if ``None``.

        """
        extent = as_tensor(extent, device=device or "cpu")
        if not extent.is_floating_point():
            extent = extent.float()
        self._extent = extent
        # Set other properties AFTER _extent, which implicitly defines 'device' and 'ndim'.
        # Use in-place setters to take care of conversion and value assertions.
        if direction is None:
            direction = torch.eye(self.ndim, dtype=self.dtype, device=self.device)
        self.direction_(direction)
        # Set center to default, specified center point, or derived from origin
        if origin is None:
            self.center_(0 if center is None else center)
        elif center is None:
            # ATTENTION: This must be done AFTER extent and direction are set!
            self.origin_(origin)
        else:
            self.center_(center)
            if not torch.allclose(origin, self.origin()):
                raise ValueError("Cube() 'center' and 'origin' are inconsistent")

    def numpy(self) -> np.ndarray:
        r"""Get cube attributes as 1-dimensional NumPy array."""
        return np.concatenate(
            [
                self._extent.numpy(),
                self._center.numpy(),
                self._direction.flatten().numpy(),
            ],
            axis=0,
        )

    @classmethod
    def from_numpy(cls, attrs: Union[Sequence[float], np.ndarray], origin: bool = False) -> Cube:
        r"""Create Cube from 1-dimensional NumPy array."""
        if isinstance(attrs, np.ndarray):
            seq = attrs.astype(float).tolist()
        else:
            seq = attrs
        return cls.from_seq(seq, origin=origin)

    @classmethod
    def from_seq(cls, attrs: Sequence[float], origin: bool = False) -> Cube:
        r"""Create Cube from sequence of attribute values.

        Args:
            attrs: Array of length (D + 2) * D, where ``D=2`` or ``D=3`` is the number
                of spatial cube dimensions and array items are given as
                ``(sx, ..., cx, ..., d11, ..., d21, ....)``, where ``(sx, ...)`` is the
                cube extent, ``(cx, ...)`` the cube center coordinates, and ``(d11, ...)``
                are the cube direction cosines. The argument can be a Python list or tuple,
                NumPy array, or PyTorch tensor.
            origin: Whether ``(cx, ...)`` specifies Cube origin rather than center.

        Returns:
            Cube instance.

        """
        if len(attrs) == 8:
            d = 2
        elif len(attrs) == 15:
            d = 3
        else:
            raise ValueError(
                f"{cls.__name__}.from_seq() expected array of length 8 (D=2) or 15 (D=3)"
            )
        kwargs = dict(
            extent=attrs[0:d],
            direction=attrs[2 * d :],
        )
        if origin:
            kwargs["origin"] = attrs[d : 2 * d]
        else:
            kwargs["center"] = attrs[d : 2 * d]
        return Cube(**kwargs)

    @classmethod
    def from_grid(cls, grid: Grid, align_corners: Optional[bool] = None) -> Cube:
        r"""Get cube with respect to which normalized grid coordinates are defined."""
        if align_corners is not None:
            grid = grid.align_corners(align_corners)
        return cls(
            extent=grid.cube_extent(),
            center=grid.center(),
            direction=grid.direction(),
            device=grid.device,
        )

    def grid(
        self,
        size: Optional[Union[int, Size, Array]] = None,
        shape: Optional[Union[int, Shape, Array]] = None,
        spacing: Optional[Union[Array, float]] = None,
        align_corners: bool = ALIGN_CORNERS,
    ) -> Grid:
        r"""Create regular sampling grid which covers the world space bounded by the cube."""
        if size is None and shape is None:
            if spacing is None:
                raise ValueError(
                    "Cube.grid() requires either the desired grid 'size'/'shape' or point 'spacing'"
                )
            size = self.extent().div(spacing).round()
            size = torch.Size(size.type(torch.int).tolist())
            if align_corners:
                size = torch.Size(n + 1 for n in size)
            spacing = None
        else:
            if isinstance(size, int):
                size = (size,) * self.ndim
            if isinstance(shape, int):
                shape = (shape,) * self.ndim
            size = Grid(size=size, shape=shape).size()
        ncells = torch.tensor(size)
        if align_corners:
            ncells = ncells.sub_(1)
        ncells = ncells.to(dtype=self.dtype, device=self.device)
        grid = Grid(
            size=size,
            spacing=self.extent().div(ncells),
            center=self.center(),
            direction=self.direction(),
            align_corners=align_corners,
            device=self.device,
        )
        with torch.no_grad():
            if not torch.allclose(grid.cube_extent(), self.extent()):
                raise ValueError(
                    "Cube.grid() 'size'/'shape' times 'spacing' does not match cube extent"
                )
        return grid

    def dim(self) -> int:
        r"""Number of cube dimensions."""
        return len(self._extent)

    @property
    def ndim(self) -> int:
        r"""Number of cube dimensions."""
        return len(self._extent)

    @property
    def dtype(self) -> torch.dtype:
        r"""Get data type of cube attribute tensors."""
        return self._extent.dtype

    @property
    def device(self) -> Device:
        r"""Get device on which cube attribute tensors are stored."""
        return self._extent.device

    def clone(self) -> Cube:
        r"""Make deep copy of this instance."""
        cube = shallow_copy(self)
        for name in self.__slots__:
            value = getattr(self, name)
            if isinstance(value, Tensor):
                setattr(cube, name, value.clone())
        return cube

    def __deepcopy__(self, memo) -> Cube:
        r"""Support copy.deepcopy to clone this cube."""
        if id(self) in memo:
            return memo[id(self)]
        copy = self.clone()
        memo[id(self)] = copy
        return copy

    @overload
    def center(self) -> Tensor:
        r"""Get center point in world space."""
        ...

    @overload
    def center(self, arg: Union[float, Array], *args: float) -> Cube:
        r"""Get new cube with same orientation and extent, but specified center point."""
        ...

    def center(self, *args) -> Union[Tensor, Cube]:
        r"""Get center point in world space or new cube with specified center point."""
        if args:
            return shallow_copy(self).center_(*args)
        return self._center

    def center_(self, arg: Union[Array, float], *args: float) -> Cube:
        r"""Set center point in world space of this cube."""
        self._center = cat_scalars(arg, *args, num=self.ndim, dtype=self.dtype, device=self.device)
        return self

    @overload
    def origin(self) -> Tensor:
        r"""Get world coordinates of lower left corner."""
        ...

    @overload
    def origin(self, arg: Union[Array, float], *args: float) -> Cube:
        r"""Get new cube with specified world coordinates of lower left corner."""
        ...

    def origin(self, *args) -> Union[Tensor, Cube]:
        r"""Get origin in world space or new cube with specified origin."""
        if args:
            return shallow_copy(self).origin_(*args)
        offset = torch.matmul(self.direction(), self.spacing())
        origin = self._center.sub(offset)
        return origin

    def origin_(self, arg: Union[Array, float], *args: float) -> Cube:
        r"""Set world coordinates of lower left corner."""
        center = cat_scalars(arg, *args, num=self.ndim, dtype=self.dtype, device=self.device)
        offset = torch.matmul(self.direction(), self.spacing())
        self._center = center.add(offset)
        return self

    def spacing(self) -> Tensor:
        r"""Cube unit spacing in world space."""
        return self._extent.div(2)

    @overload
    def direction(self) -> Tensor:
        r"""Get edge direction cosines matrix."""
        ...

    @overload
    def direction(self, arg: Union[Array, float], *args: float) -> Cube:
        r"""Get new cube with specified edge direction cosines."""
        ...

    def direction(self, *args) -> Union[Tensor, Cube]:
        r"""Get edge direction cosines matrix or new cube with specified orientation."""
        if args:
            return shallow_copy(self).direction_(*args)
        return self._direction

    def direction_(self, arg: Union[Array, float], *args: float) -> Cube:
        r"""Set edge direction cosines matrix of this cube."""
        D = self.ndim
        if args:
            direction = torch.tensor((arg,) + args)
        else:
            direction = as_tensor(arg)
        direction = direction.to(dtype=self.dtype, device=self.device)
        if direction.ndim == 1:
            if direction.shape[0] != D * D:
                raise ValueError(
                    f"Cube direction must be array or square matrix with numel={D * D}"
                )
            direction = direction.reshape(D, D)
        else:
            if (
                direction.ndim != 2
                or direction.shape[0] != direction.shape[1]
                or direction.shape[0] != D
            ):
                raise ValueError(
                    f"Cube direction must be array or square matrix with numel={D * D}"
                )
        with torch.no_grad():
            if abs(direction.det().abs().item() - 1) > 1e-4:
                raise ValueError("Cube direction cosines matrix must be valid rotation matrix")
        self._direction = direction
        return self

    @overload
    def extent(self) -> Tensor:
        r"""Extent of cube in world space."""
        ...

    @overload
    def extent(self, arg: Union[float, Array], *args, float) -> Cube:
        r"""Get cube with same center and orientation but different extent."""
        ...

    def extent(self, *args) -> Union[Tensor, Cube]:
        r"""Get extent of this cube or a new cube with same center and orientation but specified extent."""
        if args:
            return shallow_copy(self).extent_(*args)
        return self._extent

    def extent_(self, arg: Union[Array, float], *args) -> Cube:
        r"""Set the extent of this cube, keeping center and orientation the same."""
        self._extent = cat_scalars(arg, *args, num=self.ndim, device=self.device, dtype=self.dtype)
        return self

    def affine(self) -> Tensor:
        r"""Affine transformation from cube to world space, excluding translation."""
        return torch.mm(self.direction(), torch.diag(self.spacing()))

    def inverse_affine(self) -> Tensor:
        r"""Affine transformation from world to cube space, excluding translation."""
        one = torch.tensor(1, dtype=self.dtype, device=self.device)
        return torch.mm(torch.diag(one / self.spacing()), self.direction().t())

    def transform(
        self,
        axes: Optional[Union[Axes, str]] = None,
        to_axes: Optional[Union[Axes, str]] = None,
        to_cube: Optional[Cube] = None,
        vectors: bool = False,
    ) -> Tensor:
        r"""Transformation of coordinates from this cube to another cube.

        Args:
            axes: Axes with respect to which input coordinates are defined.
                If ``None`` and also ``to_axes`` and ``to_cube`` is ``None``,
                returns the transform which maps from cube to world space.
            to_axes: Axes of cube to which coordinates are mapped. Use ``axes`` if ``None``.
            to_cube: Other cube. Use ``self`` if ``None``.
            vectors: Whether transformation is used to rescale and reorient vectors.

        Returns:
            If ``vectors=False``, a homogeneous coordinate transformation of shape ``(D, D + 1)``.
            Otherwise, a square transformation matrix of shape ``(D, D)`` is returned.

        """
        if axes is None and to_axes is None and to_cube is None:
            return self.transform(Axes.CUBE, Axes.WORLD, vectors=vectors)
        if axes is None:
            raise ValueError(
                "Cube.transform() 'axes' required when 'to_axes' or 'to_cube' specified"
            )
        axes = Axes(axes)
        to_axes = axes if to_axes is None else Axes(to_axes)
        if axes is Axes.GRID or to_axes is Axes.GRID:
            raise ValueError("Cube.transform() Axes.GRID is only valid for a Grid")
        if axes == to_axes and axes is Axes.CUBE_CORNERS:
            axes = to_axes = Axes.CUBE
        elif axes is Axes.CUBE_CORNERS and to_axes is Axes.WORLD:
            axes = Axes.CUBE
        elif axes is Axes.WORLD and to_axes is Axes.CUBE_CORNERS:
            to_axes = Axes.CUBE
        if axes is Axes.CUBE_CORNERS or to_axes is Axes.CUBE_CORNERS:
            raise ValueError(
                "Cube.transform() cannot map between Axes.CUBE and Axes.CUBE_CORNERS."
                " Use Cube.grid().transform() instead."
            )
        if axes == to_axes and (axes is Axes.WORLD or to_cube is None or to_cube == self):
            return torch.eye(self.ndim, dtype=self.dtype, device=self.device)
        if axes == to_axes:
            assert axes is Axes.CUBE
            cube_to_world = self.transform(Axes.CUBE, Axes.WORLD, vectors=vectors)
            world_to_cube = to_cube.transform(Axes.WORLD, Axes.CUBE, vectors=vectors)
            if vectors:
                return torch.mm(world_to_cube, cube_to_world)
            return hmm(world_to_cube, cube_to_world)
        if axes is Axes.CUBE:
            assert to_axes is Axes.WORLD
            if vectors:
                return self.affine()
            return homogeneous_matrix(self.affine(), self.center())
        assert axes is Axes.WORLD
        assert to_axes is Axes.CUBE
        if vectors:
            return self.inverse_affine()
        return hmm(self.inverse_affine(), -self.center())

    def inverse_transform(self, vectors: bool = False) -> Tensor:
        r"""Transform which maps from world to cube space."""
        return self.transform(Axes.WORLD, Axes.CUBE, vectors=vectors)

    def apply_transform(
        self,
        arg: Array,
        axes: Union[Axes, str],
        to_axes: Optional[Union[Axes, str]] = None,
        to_cube: Optional[Cube] = None,
        vectors: bool = False,
    ) -> Tensor:
        r"""Map point coordinates or displacement vectors from one cube to another.

        Args:
            arg: Coordinates of points or displacement vectors as tensor of shape ``(..., D)``.
            axes: Axes of this cube with respect to which input coordinates are defined.
            to_axes: Axes of cube to which coordinates are mapped. Use ``axes`` if ``None``.
            to_cube: Other cube. Use ``self`` if ``None``.
            vectors: Whether ``arg`` contains displacements (``True``) or point coordinates (``False``).

        Returns:
            Points or displacements with respect to ``to_cube`` and ``to_axes``.
            If ``to_cube == self`` and ``to_axes == axes`` or both ``axes`` and ``to_axes`` are
            ``Axes.WORLD`` and ``arg`` is a ``torch.Tensor``, a reference to the unmodified input
            tensor is returned.

        """
        axes = Axes(axes)
        to_axes = axes if to_axes is None else Axes(to_axes)
        if to_cube is None:
            to_cube = self
        tensor = as_tensor(arg)
        if not tensor.is_floating_point():
            tensor = tensor.type(self.dtype)
        if axes is to_axes and axes is Axes.WORLD:
            return tensor
        if (to_cube is not None and to_cube != self) or axes is not to_axes:
            matrix = self.transform(axes, to_axes, to_cube=to_cube, vectors=vectors)
            matrix = matrix.unsqueeze(0).to(device=tensor.device)
            tensor = homogeneous_transform(matrix, tensor)
        return tensor

    def transform_points(
        self,
        points: Array,
        axes: Union[Axes, str],
        to_axes: Optional[Union[Axes, str]] = None,
        to_cube: Optional[Cube] = None,
    ) -> Tensor:
        r"""Map point coordinates from one cube to another.

        Args:
            points: Coordinates of points to transform as tensor of shape ``(..., D)``.
            axes: Axes of this cube with respect to which input coordinates are defined.
            to_axes: Axes of cube to which coordinates are mapped. Use ``axes`` if ``None``.
            to_cube: Other cube. Use ``self`` if ``None``.

        Returns:
            Point coordinates with respect to ``to_cube`` and ``to_axes``. If ``to_cube == self``
            and ``to_axes == axes`` or both ``axes`` and ``to_axes`` are ``Axes.WORLD`` and ``arg``
            is a ``torch.Tensor``, a reference to the unmodified input tensor is returned.

        """
        return self.apply_transform(points, axes, to_axes, to_cube=to_cube, vectors=False)

    def transform_vectors(
        self,
        vectors: Array,
        axes: Union[Axes, str],
        to_axes: Optional[Union[Axes, str]] = None,
        to_cube: Optional[Cube] = None,
    ) -> Tensor:
        r"""Rescale and reorient flow vectors.

        Args:
            vectors: Displacement vectors as tensor of shape ``(..., D)``.
            axes: Axes of this cube with respect to which input coordinates are defined.
            to_axes: Axes of cube to which coordinates are mapped. Use ``axes`` if ``None``.
            to_cube: Other cube. Use ``self`` if ``None``.

        Returns:
            Rescaled and reoriented displacement vectors. If ``to_cube == self`` and
            ``to_axes == axes`` or both ``axes`` and ``to_axes`` are ``Axes.WORLD`` and ``arg``
            is a ``torch.Tensor``, a reference to the unmodified input tensor is returned.

        """
        return self.apply_transform(vectors, axes, to_axes, to_cube=to_cube, vectors=True)

    def cube_to_world(self, coords: Array) -> Tensor:
        r"""Map point coordinates from cube to world space.

        Args:
            coords: Normalized coordinates with respect to this cube as tensor of shape ``(..., D)``.

        Returns:
            Coordinates of points in world space.

        """
        return self.apply_transform(coords, Axes.CUBE, Axes.WORLD, vectors=False)

    def world_to_cube(self, points: Array) -> Tensor:
        r"""Map point coordinates from world to cube space.

        Args:
            points: Coordinates of points in world space as tensor of shape ``(..., D)``.

        Returns:
            Normalized coordinates of points with respect to this cube.

        """
        return self.apply_transform(points, Axes.WORLD, Axes.CUBE, vectors=False)

    def __eq__(self, other: Any) -> bool:
        r"""Compare this cube to another."""
        if other is self:
            return True
        if not isinstance(other, self.__class__):
            return False
        for name in self.__slots__:
            value = getattr(self, name)
            other_value = getattr(other, name)
            if type(value) != type(other_value):
                return False
            if isinstance(value, Tensor):
                assert isinstance(other_value, Tensor)
                if value.shape != other_value.shape:
                    return False
                other_value = other_value.to(device=value.device)
                if not torch.allclose(value, other_value, rtol=1e-5, atol=1e-8):
                    return False
            elif value != other_value:
                return False
        return True

    def __repr__(self) -> str:
        """String representation."""
        origin = ", ".join([f"{v:.5f}" for v in self.origin()])
        center = ", ".join([f"{v:.5f}" for v in self._center])
        direction = ", ".join([f"{v:.5f}" for v in self._direction.flatten()])
        extent = ", ".join([f"{v:.5f}" for v in self._extent])
        return (
            f"{type(self).__name__}("
            + f"origin=({origin})"
            + f", center=({center})"
            + f", extent=({extent})"
            + f", direction=({direction})"
            + f", device={repr(str(self.device))}"
            + ")"
        )


def cube_points_transform(cube: Cube, axes: Axes, to_cube: Cube, to_axes: Optional[Axes] = None):
    r"""Get linear transformation of points from one cube to another.

    Args:
        cube: Sampling grid with respect to which input points are defined.
        axes: Grid axes with respect to which input points are defined.
        to_cube: Sampling grid with respect to which output points are defined.
        to_axes: Grid axes with respect to which output points are defined.

    Returns:
        Homogeneous coordinate transformation matrix as tensor of shape ``(D, D + 1)``.

    """
    return cube.transform(axes=axes, to_axes=to_axes, to_cube=to_cube, vectors=False)


def cube_vectors_transform(cube: Cube, axes: Axes, to_cube: Cube, to_axes: Optional[Axes] = None):
    r"""Get affine transformation which maps vectors with respect to one cube to another.

    Args:
        cube: Cube with respect to which (normalized) input vectors are defined.
        axes: Cube axes with respect to which input vectors are defined.
        to_cube: Cube with respect to which (normalized) output vectors are defined.
        to_axes: Cube axes with respect to which output vectors are defined.

    Returns:
        Affine transformation matrix as tensor of shape ``(D, D)``.

    """
    return cube.transform(axes=axes, to_axes=to_axes, to_cube=to_cube, vectors=True)


def cube_transform_points(
    points: Tensor,
    cube: Cube,
    axes: Axes,
    to_cube: Cube,
    to_axes: Optional[Axes] = None,
):
    return cube.transform_points(points, axes=axes, to_axes=to_axes, to_cube=to_cube)


def cube_transform_vectors(
    vectors: Tensor,
    cube: Cube,
    axes: Axes,
    to_cube: Cube,
    to_axes: Optional[Axes] = None,
):
    return cube.transform_vectors(vectors, axes=axes, to_axes=to_axes, to_cube=to_cube)



r"""Auxiliary functions for conversion between SimpleITK and PyTorch."""

from typing import Optional, Sequence

import SimpleITK as sitk
import torch


def image_from_tensor(
    data: torch.Tensor,
    origin: Optional[Sequence[float]] = None,
    spacing: Optional[Sequence[float]] = None,
    direction: Optional[Sequence[float]] = None,
) -> sitk.Image:
    r"""Create ``SimpleITK.Image`` from image data tensor.

    Args:
        data: Image tensor of shape ``(C, ..., X)``.
        origin: World coordinates of center of voxel with zero indices.
        spacing: Voxel size in world units in each dimension.
        direction: Flattened image orientation cosine matrix.

    Returns:
        SimpleITK image.

    """
    data = data.detach().cpu()
    nchannels = data.shape[0]
    if nchannels == 1:
        data = data[0]
    else:
        data = data.unsqueeze(-1).transpose(0, -1).squeeze(0)
    image = sitk.GetImageFromArray(data.numpy(), isVector=nchannels > 1)
    if origin:
        image.SetOrigin(origin)
    if spacing:
        image.SetSpacing(spacing)
    if direction:
        image.SetDirection(direction)
    return image


def tensor_from_image(
    image: sitk.Image,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
) -> torch.Tensor:
    r"""Create image data tensor from ``SimpleITK.Image``."""
    if image.GetPixelID() == sitk.sitkUInt16:
        image = sitk.Cast(image, sitk.sitkInt32)
    elif image.GetPixelID() == sitk.sitkUInt32:
        image = sitk.Cast(image, sitk.sitkInt64)
    data = torch.from_numpy(sitk.GetArrayFromImage(image))
    data = data.unsqueeze(0)
    if image.GetNumberOfComponentsPerPixel() > 1:
        data = data.transpose(0, -1).squeeze(-1)
    return data.to(dtype=dtype, device=device)



r"""Bridge between VTK and NumPy."""

from typing import Optional, Union
import warnings

import numpy as np

from vtk import (
    VTK_CHAR,
    VTK_SHORT,
    VTK_LONG,
    VTK_LONG_LONG,
    VTK_UNSIGNED_CHAR,
    VTK_UNSIGNED_SHORT,
    VTK_UNSIGNED_LONG,
    VTK_UNSIGNED_LONG_LONG,
    VTK_FLOAT,
    VTK_DOUBLE,
    vtkCellArray,
    vtkDataArray,
    vtkMatrix4x4,
    vtkPoints,
    vtkPointSet,
)
from vtk.util import numpy_support  # type: ignore


VTK_DATA_TYPE_FROM_NUMPY_DTYPE = {
    np.dtype("int8"): VTK_CHAR,
    np.dtype("int16"): VTK_SHORT,
    np.dtype("int32"): VTK_LONG,
    np.dtype("int64"): VTK_LONG_LONG,
    np.dtype("uint8"): VTK_UNSIGNED_CHAR,
    np.dtype("uint16"): VTK_UNSIGNED_SHORT,
    np.dtype("uint32"): VTK_UNSIGNED_LONG,
    np.dtype("uint64"): VTK_UNSIGNED_LONG_LONG,
    np.dtype("float32"): VTK_FLOAT,
    np.dtype("float64"): VTK_DOUBLE,
}


NUMPY_DTYPE_FROM_VTK_DATA_TYPE = {
    VTK_CHAR: np.dtype("int8"),
    VTK_SHORT: np.dtype("int16"),
    VTK_LONG: np.dtype("int32"),
    VTK_LONG_LONG: np.dtype("int64"),
    VTK_UNSIGNED_CHAR: np.dtype("uint8"),
    VTK_UNSIGNED_SHORT: np.dtype("uint16"),
    VTK_UNSIGNED_LONG: np.dtype("uint32"),
    VTK_UNSIGNED_LONG_LONG: np.dtype("uint64"),
    VTK_FLOAT: np.dtype("float32"),
    VTK_DOUBLE: np.dtype("float64"),
}


def numpy_to_vtk_matrix4x4(arr: np.ndarray) -> vtkMatrix4x4:
    """Create vtkMatrix4x4 from NumPy array."""
    assert arr.shape == (4, 4)
    matrix = vtkMatrix4x4()
    for i in range(4):
        for j in range(4):
            matrix.SetElement(i, j, arr[i, j])
    return matrix


def numpy_to_vtk_array(*args, name: Optional[str] = None, **kwargs) -> vtkDataArray:
    r"""Convert NumPy array to vtkDataArray."""
    with warnings.catch_warnings():
        # vtk/util/numpy_support.py:137: FutureWarning: Conversion of the second argument of
        # issubdtype from `complex` to `np.complexfloating` is deprecated. In future, it will
        # be treated as `np.complex128 == np.dtype(complex).type`.
        warnings.filterwarnings("ignore", category=FutureWarning)
        data_array = numpy_support.numpy_to_vtk(*args, **kwargs)
    if name is not None:
        data_array.SetName(name)
    return data_array


def numpy_to_vtk_points(arr: np.ndarray) -> vtkPoints:
    r"""Convert NumPy array to vtkPoints."""
    data = numpy_to_vtk_array(arr)
    points = vtkPoints()
    points.SetData(data)
    return points


def numpy_to_vtk_cell_array(arr: np.ndarray) -> vtkCellArray:
    r"""Convert NumPy array to vtkCellArray."""
    if arr.ndim != 2:
        raise ValueError("numpy_to_vtk_cell_array() 'arr' must be 2-dimensional")
    if arr.dtype not in (np.dtype("int32"), np.dtype("int64")):
        raise TypeError("numpy_to_vtk_cell_array() 'arr' must have dtype int32 or int64")
    cells = vtkCellArray()
    use_set_data = False
    if use_set_data:
        # FIXME: causes segfault when loading
        dtype = arr.dtype
        offsets = np.repeat(np.array(arr.shape[1], dtype=dtype), arr.shape[0])
        offsets = np.concatenate([[0], np.cumsum(offsets)])
        offsets = numpy_to_vtk_array(offsets.astype(dtype))
        connectivity = numpy_to_vtk_array(arr.flatten().astype(dtype))
        if not cells.SetData(offsets, connectivity):
            raise RuntimeError(
                "numpy_to_vtk_cell_array() failed to convert NumPy array to vtkCellArray"
            )
    else:
        cells.AllocateExact(arr.shape[0], np.prod(arr.shape))
        for cell in arr:
            cells.InsertNextCell(len(cell))
            for ptId in cell:
                cells.InsertCellPoint(ptId)
    assert cells.IsValid()
    return cells


def vtk_to_numpy_array(data: vtkDataArray) -> np.ndarray:
    r"""Convert vtkDataArray to NumPy array."""
    return numpy_support.vtk_to_numpy(data)


def vtk_to_numpy_points(points: Union[vtkPoints, vtkPointSet]) -> np.ndarray:
    r"""Convert vtkPoints to NumPy array."""
    if isinstance(points, vtkPointSet):
        points = points.GetPoints()
    return vtk_to_numpy_array(points.GetData())



r"""Base classes of spatial coordinate transformations."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from copy import copy as shallow_copy
from typing import Optional, Tuple, TypeVar, Union, final, overload

import torch
import torch.optim
from torch import Tensor
from torch.nn import Module
import torch.nn.functional as F

from deepali.core import functional as U
from deepali.core.grid import Axes, Grid
from deepali.core.linalg import as_homogeneous_matrix
from deepali.core.typing import Device
from deepali.data.flow import FlowFields
from deepali.modules import DeviceProperty


TSpatialTransform = TypeVar("TSpatialTransform", bound="SpatialTransform")
TLinearTransform = TypeVar("TLinearTransform", bound="LinearTransform")
TNonRigidTransform = TypeVar("TNonRigidTransform", bound="NonRigidTransform")


class ReadOnlyParameters(RuntimeError):
    r"""Exception thrown when attempting to set parameters when these are provided by a callable."""
    ...


class SpatialTransform(DeviceProperty, Module, metaclass=ABCMeta):
    r"""Base class of all spatial coordinate transformations."""

    def __init__(self, grid: Grid):
        r"""Initialize base class.

        Args:
            grid: Spatial domain with respect to which transformation is defined.
                The unit cube domain is :attr:`.Axes.CUBE` if ``grid.align_corners() == False``,
                and :attr:`.Axes.CUBE_CORNERS` otherwise.

        """
        if not isinstance(grid, Grid):
            raise TypeError("SpatialTransform() 'grid' must be of type Grid")
        super().__init__()
        self._grid = grid
        self._args = ()
        self._kwargs = {}
        self.register_update_hook()

    def __copy__(self: TSpatialTransform) -> TSpatialTransform:
        r"""Make shallow copy of this transformation.

        The copy shares containers for parameters and hooks with this module, but not containers of
        buffers and modules. References to currently set buffers and modules are however copied, but
        adding/removing a buffer or module to/from the shallow copy will not modify the buffers and
        modules of the original module. The same is the case for adding/removing buffers or modules
        to/from the original module.

        Returns:
            Shallow copy of this spatial transformation module.

        """
        copy = self.__new__(type(self))
        copy.__dict__ = self.__dict__.copy()
        for name in ("_buffers", "_non_persistent_buffers_set", "_modules"):
            if name in self.__dict__:
                copy.__dict__[name] = self.__dict__[name].copy()
        return copy

    @overload
    def condition(self) -> Tuple[tuple, dict]:
        r"""Get arguments on which transformation is conditioned.

        Returns:
            args: Positional arguments.
            kwargs: Keyword arguments.

        """
        ...

    @overload
    def condition(self: TSpatialTransform, *args, **kwargs) -> TSpatialTransform:
        r"""Get new transformation which is conditioned on the specified arguments."""
        ...

    def condition(
        self: TSpatialTransform, *args, **kwargs
    ) -> Union[TSpatialTransform, Tuple[tuple, dict]]:
        r"""Get or set data tensor on which transformation is conditioned."""
        if args:
            return shallow_copy(self).condition_(*args)
        return self._args, self._kwargs

    def condition_(self: TSpatialTransform, *args, **kwargs) -> TSpatialTransform:
        r"""Set data tensor on which this transformation is conditioned."""
        self.clear_buffers()
        self._args = args
        self._kwargs = kwargs
        return self

    def align_corners(self) -> bool:
        r"""Whether extrema -1 and 1 coincide with grid border (False) or corner points (True)."""
        return self._grid.align_corners()

    def axes(self) -> Axes:
        """Axes with respect to which transformation is defined.

        Returns:
            ``Axes.CUBE_CORNERS if self.align_corners() else Axes.CUBE``.

        """
        return Axes.from_align_corners(self.align_corners())

    @overload
    def grid(self) -> Grid:
        ...

    @overload
    def grid(self: TSpatialTransform, grid: Grid) -> TSpatialTransform:
        ...

    def grid(self, grid: Optional[Grid] = None) -> Grid:
        r"""Get grid domain of this transformation or a new transformation with the specified grid."""
        if grid is None:
            return self._grid
        return shallow_copy(self).grid_(grid)

    def grid_(self: TSpatialTransform, grid: Grid) -> TSpatialTransform:
        r"""Set sampling grid which defines domain and codomain of this transformation."""
        if self._grid == grid:
            return self
        if grid.ndim != self.ndim:
            raise ValueError(f"{type(self).__name__}.grid_() must be {self.ndim}-dimensional")
        self.clear_buffers()
        self._grid = grid
        return self

    def dim(self) -> int:
        r"""Number of spatial dimensions."""
        return self._grid.ndim

    @property
    def ndim(self) -> int:
        r"""Number of spatial dimensions."""
        return self.dim()

    @property
    def linear(self) -> bool:
        r"""Whether this transformation is linear."""
        return isinstance(self, LinearTransform)

    @property
    def nonrigid(self) -> bool:
        r"""Whether this transformation is non-rigid."""
        return not self.linear

    def fit(self: TSpatialTransform, flow: FlowFields, **kwargs) -> TSpatialTransform:
        r"""Fit transformation to a given flow field.

        Args:
            flow: Flow fields to approximate.
            kwargs: Optional keyword arguments of fitting algorithm.
                Arguments which are unused by a concrete implementation are ignored
                without raising an error, e.g., a specified gradient descent step length
                when a least square fit is computed instead.

        Returns:
            Reference to this transformation.

        Raises:
            RuntimeError: When this transformation has no optimizable parameters.

        """
        self.clear_buffers()
        grid = self.grid()
        flow = flow.to(self.device)
        flow = flow.sample(grid)
        flow = flow.axes(Axes.from_grid(grid))
        self._fit(flow, **kwargs)
        return self

    def _fit(self, flow: FlowFields, **kwargs) -> None:
        r"""Fit transformation to flow field.

        This function may be overidden by subclasses to implement an analytic least squares
        or otherwise more suitable fitting approach. Optimization related keyword arguments
        may be ignored by these specializations.

        Args:
            flow: Batch of flow vector fields sampled on ``self.grid()`` and
                defined with respect to either ``Axes.CUBE`` or ``Axes.CUBE_CORNERS``
                depending on flag ``self.grid().align_corners()``. These displacement vector
                fields will be approximated by this transformation.
            kwargs: Keyword arguments of iterative optimization. Unused arguments are ignored.
                lr: Initial step size for iterative gradient-based optimization.
                steps: Maximum number of gradient steps after which to terminate.
                epsilon: Upper mean squared error threshold at which to terminate.

        Raises:
            RuntimeError: When this transformation has no optimizable parameters.

        """
        lr = float(kwargs.get("lr", 0.1))
        steps = int(kwargs.get("steps", 1000))
        epsilon = float(kwargs.get("epsilon", 1e-5))
        verbose = int(kwargs.get("verbose", 0))
        params = list(self.parameters())
        if not params:
            raise RuntimeError(
                f"{type(self).__name__}.fit() transformation has no optimizable parameters"
            )
        optimizer = torch.optim.Adam(params, lr=lr)
        for step in range(steps):
            optimizer.zero_grad()
            loss = F.mse_loss(self.disp(), flow.tensor())
            loss.backward()
            optimizer.step()
            error = loss.detach()
            converged = error.le(epsilon).all()
            if verbose > 0 and (converged or step % verbose == 0):
                print(f"{type(self).__name__}.fit(): step={step}, mse={error.tolist()}")
            if converged:
                break

    def forward(self, points: Tensor, grid: bool = False) -> Tensor:
        r"""Transform normalized points by this spatial transformation.

        Args:
            points: Tensor of shape ``(N, M, D)`` or ``(N, ..., Y, X, D)``.
            grid: Whether ``points`` are the positions of undeformed grid points.

        Returns:
            Tensor of same shape as ``points`` with transformed point coordinates.

        """
        transform = self.tensor().to(points.device)
        apply = U.transform_grid if grid else U.transform_points
        align_corners = self.align_corners()
        return apply(transform, points, align_corners=align_corners)

    def points(
        self,
        points: Tensor,
        grid: Optional[Grid] = None,
        axes: Optional[Union[Axes, str]] = None,
        to_grid: Optional[Grid] = None,
        to_axes: Optional[Union[Axes, str]] = None,
    ) -> Tensor:
        r"""Transform points by this spatial transformation.

        Args:
            points: Tensor of shape ``(N, M, D)`` or ``(N, ..., Y, X, D)``.
            grid: Grid with respect to which input ``points`` are defined. Uses ``self.grid()`` if ``None``.
            axes: Coordinate axes with respect to which ``points`` are defined. Uses ``self.axes()`` if ``None``.
            to_grid: Grid with respect to which output points are defined. Same as ``grid`` if ``None``.
            to_axes: Coordinate axes to which ``points`` should be mapped to. Same as ``axes`` if ``None``.

        Returns:
            Point coordinates in ``(grid, axes)`` spatially transformed and mapped to coordinates with respect to ``(to_grid, to_axes)``.

        """
        if grid is None:
            grid = self.grid()
        if axes is None:
            axes = self.axes()
        else:
            axes = Axes.from_arg(axes)
        if to_grid is None:
            to_grid = grid
        if to_axes is None:
            to_axes = axes
        else:
            to_axes = Axes.from_arg(to_axes)
        points = grid.transform_points(
            points,
            axes=axes,
            to_grid=self.grid(),
            to_axes=self.axes(),
            decimals=None,
        )
        points = self.forward(points)
        points = self.grid().transform_points(
            points,
            axes=self.axes(),
            to_grid=to_grid,
            to_axes=to_axes,
            decimals=None,
        )
        return points

    def disp(self, grid: Optional[Grid] = None) -> Tensor:
        r"""Get displacement vector field representation of this transformation.

        Args:
            grid: Grid on which to sample vector fields. Use ``self.grid()`` if ``None``.

        Returns:
            Displacement vector fields as tensor of shape ``(N, D, ..., X)``.

        """
        if grid is None:
            grid = self.grid()
        data = self.tensor()
        if data.ndim < 3:
            raise AssertionError(
                f"SpatialTransform.disp() expected {type(self).__name__}.tensor()"
                f" to be at least 3-dimensional"
            )
        if data.shape[1] != self.dim():
            raise AssertionError(
                f"SpatialTransform.disp() expected {type(self).__name__}.tensor()"
                f" shape[1] to be equal to {type(self).__name__}.dim()={self.dim()}"
            )
        # Affine transformation with shape:
        # - (N, D, 1): Translation only.
        # - (N, D, D): Affine only, i.e., no translation.
        # - (N, D, D + 1): Affine transformation, including translation.
        if data.ndim == 3:
            assert self.linear
            data = U.affine_flow(data, grid)
        # Non-rigid deformation tensor as displacement field with shape (N, D, ..., X)
        else:
            assert not self.linear
            align_corners = grid.align_corners()
            # Displacement field with domain different from output domain
            # - Use F.grid_sample() to resample displacement field and adjust vectors.
            if grid != self.grid() or align_corners != self.align_corners():
                flow = FlowFields(data, grid=self.grid().reshape(data.shape[2:]))
                flow = flow.sample(grid)
                data = flow.tensor()
            # Displacement field with same domain as output grid, but differing size
            # - Use F.interpolate() to resize displacement field.
            elif grid.shape != data.shape[2:]:
                data = U.grid_reshape(data, grid.shape, align_corners=align_corners)
        return data

    @final
    def flow(self, grid: Optional[Grid] = None, device: Optional[Device] = None) -> FlowFields:
        r"""Get flow field representation of this transformation.

        Args:
            grid: Grid on which to sample flow fields. Use ``self.grid()`` if ``None``.
            device: Device on which to store returned flow fields.

        Returns:
            Flow fields defined by a tensor of shape ``(N, D, ..., X)`` and a common spatial grid.

        """
        if grid is None:
            grid = self.grid()
        data = self.disp(grid)
        return FlowFields(data, grid=grid, device=device)

    @abstractmethod
    def tensor(self) -> Tensor:
        r"""Get tensor representation of this transformation.

        The tensor representation of a transformation is with respect to the unit cube axes defined
        by its sampling grid as specified by ``self.axes()``. For a non-rigid transformation it is
        a displacement vector field. For linear transformations, it is a batch of homogeneous
        transformation tensors whose shape determines the type of linear transformation.

        Returns:
            Returns a batch of homogeneous transformation matrices as tensor of shape ``(N, D, 1)``
            (translation),  ``(N, D, D)`` (affine) or ``(N, D, D + 1)``, i.e., a 3-dimensional tensor,
            if this transformation is a :class:`.LinearTransform`. In case of a non-rigid transformation,
            a displacement vector field is returned as tensor of shape ``(N, D, ..., X)``, i.e., a
            higher dimensional tensor, where ``D = self.ndim`` and the number of tensor dimensions is
            equal to ``D + 2``.

        """
        raise NotImplementedError(f"{type(self).__name__}.tensor()")

    def update(self: TSpatialTransform) -> TSpatialTransform:
        r"""Update internal state of this transformation.

        This function is called by a pre-forward hook. It can be overriden by subclasses to
        update their internal state, e.g., to obtain current predictions of transformation
        parameters, to compute a dense vector field from spline coefficients, to compute
        a displacement field from a velocity field, etc. When calling other functions than
        the module's ``__call__`` function, the ``update()`` of the transformation must be
        called explicitly unless it is known that the specific transformation keeps no
        internal state other than its (optimizable) parameters.

        """
        return self

    @staticmethod
    def _update_hook(transform: Module, *args, **kwargs) -> None:
        r"""Update callback which is registered as pre-forward hook."""
        assert isinstance(transform, SpatialTransform)
        transform.update()

    def register_update_hook(self) -> None:
        r"""Register a forward pre-hook which invokes :meth:`.SpatialTransform.update`."""
        self._update_hook_handle = self.register_forward_pre_hook(self._update_hook)

    def remove_update_hook(self) -> None:
        r"""Remove previously registered :meth:`.SpatialTransform.update` hook."""
        if self._update_hook_handle is not None:
            self._update_hook_handle.remove()
            self._update_hook_handle = None

    @property
    def inv(self: TSpatialTransform) -> TSpatialTransform:
        r"""Get inverse transformation.

        Convenience property for applying the inverse transformation, e.g.,

        .. code-block:: python

            y = transform(x)
            x = transform.inv(y)

        """
        return self.inverse(link=True, update_buffers=True)

    def inverse(
        self: TSpatialTransform, link: bool = False, update_buffers: bool = False
    ) -> TSpatialTransform:
        r"""Get inverse of this transformation.

        Args:
            link: Whether the inverse transformation keeps a reference to this transformation.
                If ``True``, the ``update()`` function of the inverse function will not recompute
                shared parameters, e.g., parameters obtained by a callable neural network, but
                directly access the parameters from this transformation. Note that when ``False``,
                the inverse transformation will still share parameters, modules, and buffers with
                this transformation, but these shared tensors may be replaced by a call of ``update()``
                (which is implicitly called as pre-forward hook when ``__call__()`` is invoked).
            update_buffers: Whether buffers of inverse transformation should be updated after creating
                the shallow copy. If ``False``, the ``update()`` function of the returned inverse
                transformation has to be called before it is used.

        Returns:
            Shallow copy of this transformation which computes and applies the inverse transformation.
            The inverse transformation will share the parameters with this transformation. Not all
            transformations may implement this functionality.

        Raises:
            NotImplementedError: Transformation does not support sharing parameters with its inverse.

        """
        raise NotImplementedError(f"{type(self).__name__}.inverse()")

    def clear_buffers(self: TSpatialTransform) -> TSpatialTransform:
        r"""Clear any buffers that are registered by ``self.update()``."""
        ...

    def extra_repr(self) -> str:
        return f"grid={repr(self.grid())}"


class LinearTransform(SpatialTransform):
    r"""Homogeneous coordinate transformation."""

    @overload
    def matrix(self) -> Tensor:
        r"""Get matrix representation of linear transformation."""
        ...

    @overload
    def matrix(self: TLinearTransform, arg: Tensor) -> TLinearTransform:
        r"""Get shallow copy of this transformation with parameters obtained from given matrix."""
        ...

    @final
    def matrix(
        self: TLinearTransform, arg: Optional[Tensor] = None
    ) -> Union[TLinearTransform, Tensor]:
        r"""Get matrix representation of linear transformation or shallow copy with parameters set from matrix."""
        if arg is None:
            return as_homogeneous_matrix(self.tensor())
        return shallow_copy(self).matrix_(arg)

    def matrix_(self: TLinearTransform, arg: Tensor) -> TLinearTransform:
        raise NotImplementedError(f"{type(self).__name__}.matrix_()")


class NonRigidTransform(SpatialTransform):
    r"""Base class for non-linear transformation models.

    All non-linear transformation models parameterize a dense displacement vector field, either a separate
    displacement for each image in an input batch (e.g., for groupwise registration, batched registration),
    or a single displacement field used to deform all images in an input batch. The parameterization, and
    thereby the set of optimizable parameters, is defined by subclasses. The ``tensor()`` function must be
    implemented by subclasses to evaluate the non-parametric dense displacement field given the current model
    parameters. The flow vectors must be with respect to the grid ``Axes.CUBE``, i.e., where coordinate
    -1 corresponds to the left edge of the unit cube with side length 2, and coordinate 1 the right edge of
    this unit cube, respectively. Note that this corresponds to option ``align_corners=False`` of
    ``torch.nn.functional.grid_sample``.

    """

    @final
    def tensor(self) -> Tensor:
        r"""Get tensor representation of this transformation.

        Returns:
            Batch of displacement vector fields as tensor of shape ``(N, D, ..., X)``.

        """
        u = getattr(self, "u", None)
        if u is None:
            u = getattr(self.update(), "u", None)
        if u is None or "u" not in {name for name, _, in self.named_buffers()}:
            raise AssertionError(
                f"{type(self).__name__}.update() required to register"
                " displacement vector field tensor as buffer named 'u'."
                " See also NonRigidTransform.update() docstring."
            )
        if not isinstance(u, Tensor):
            raise AssertionError(f"{type(self).__name__}.tensor() 'u' must be tensor")
        if u.ndim != self.ndim + 2:
            raise AssertionError(
                f"{type(self).__name__}.tensor() 'u' must be {self.ndim + 2}-dimensional"
            )
        if u.shape[1] != self.ndim:
            raise AssertionError(
                f"{type(self).__name__}.tensor() 'u' must have shape (N, {self.ndim}, ..., X)"
            )
        return u

    def update(self: TNonRigidTransform) -> TNonRigidTransform:
        r"""Update buffered vector fields.

        Required:
            ``u``: Displacement vector field representation of non-rigid transformation.

        Optional:
            ``v``: Velocity vector field representation of non-rigid transformation if applicable.
                When this buffer is set, it can be used in a regularization term to encourage smoothness
                or other desired properties on the (stationary) velocity field. Alternatively, a
                regularization term may be based directly on the optimizable parameters.

        Returns:
            Self reference to this updated transformation.

        """
        return self

    def clear_buffers(self: TNonRigidTransform) -> TNonRigidTransform:
        r"""Clear any buffers that are registered by ``self.update()``."""
        super().clear_buffers()
        for name in ("u", "v"):
            try:
                delattr(self, name)
            except AttributeError:
                pass
        return self



r"""Point set distance terms."""

import torch
import torch.linalg
from torch import Tensor

from deepali.core import functional as U

from .base import PointSetDistance


class ClosestPointDistance(PointSetDistance):
    r"""Average closest point distance."""

    def __init__(self, scale: float = 10, split_size: int = 1e5):
        r"""Initialize closest point distance loss.

        Args:
            scale: Constant factor by which to scale average closest point
                distance value such that magnitude is in similar range to
                other registration loss terms, i.e., image similarity losses.
            split_size: Number of points by which to split point sets during
                distance calculation to avoid running out of memory. For each
                split, all pairwise point distances are calculated, followed
                by a reduction of the results by selecting the minimum across
                all splits.

        """
        super().__init__()
        self.scale = float(scale)
        self.split_size = int(split_size)

    def forward(self, x: Tensor, *ys: Tensor) -> Tensor:
        r"""Evaluate point set distance."""
        if not ys:
            raise ValueError(f"{type(self).__name__}.forward() requires at least two point sets")
        x = x.float()
        loss = torch.tensor(0, dtype=x.dtype, device=x.device)
        for y in ys:
            with torch.no_grad():
                indices = U.closest_point_indices(x, y, split_size=self.split_size)
            y = U.batched_index_select(y, 1, indices)
            dists: Tensor = torch.linalg.norm(x - y, ord=2, dim=2)
            loss += dists.mean()
        return self.scale * loss / len(ys)

    def extra_repr(self) -> str:
        return f"scale={self.scale}, split_size={self.split_size}"


CPD = ClosestPointDistance


class LandmarkPointDistance(PointSetDistance):
    r"""Average distance between corresponding landmarks."""

    def __init__(self, scale: float = 10):
        r"""Initialize point distance loss.

        Args:
            scale: Constant factor by which to scale average point distance value
                such that magnitude is in similar range to other registration loss
                terms, i.e., image similarity losses.

        """
        super().__init__()
        self.scale = float(scale)

    def forward(self, x: Tensor, *ys: Tensor) -> Tensor:
        r"""Evaluate point set distance."""
        if not ys:
            raise ValueError(f"{type(self).__name__}.forward() requires at least two point sets")
        x = x.float()
        loss = torch.tensor(0, dtype=x.dtype, device=x.device)
        for y in ys:
            dists: Tensor = torch.linalg.norm(x - y, ord=2, dim=2)
            loss += dists.mean()
        return self.scale * loss / len(ys)

    def extra_repr(self) -> str:
        return f"scale={self.scale}"


LPD = LandmarkPointDistance



r"""Utilities for parsing and setting up of common environment variables."""

import os
import re
from typing import Optional, Tuple


def cuda_visible_devices() -> Tuple[int, ...]:
    r"""Get IDs of GPUs specified by CUDA_VISIBLE_DEVICES environment variable."""
    gpus = os.environ.get("CUDA_VISIBLE_DEVICES", "")
    if not gpus:
        return ()
    gpus = [x for x in gpus.split(",") if x]
    gpu_ids = []
    RE_RANGE = re.compile(r"^(?P<start>[0-9]+)-(?P<end>[0-9]+)$")
    for gpu in gpus:
        match = RE_RANGE.match(gpu)
        if match is None:
            try:
                gpu_ids.append(int(gpu))
            except TypeError:
                raise TypeError(f"CUDA_VISIBLE_DEVICES contains invalid value {gpu}")
        else:
            gpu_ids.extend(range(int(match.group("start")), int(match.group("end")) + 1))
    for gpu_id in gpu_ids:
        if gpu_id < 0:
            raise ValueError("CUDA_VISIBLE_DEVICES contains negative GPU ID")
    return gpu_ids


def check_cuda_visible_devices(num: Optional[int] = None) -> int:
    r"""Check if CUDA_VISIBLE_DEVICES environment variable is set."""
    gpu_ids = cuda_visible_devices()
    if num and len(gpu_ids) != num:
        raise RuntimeError(f"CUDA_VISIBLE_DEVICES must be set to {num} GPUs")
    return len(gpu_ids)


def init_num_threads(threads: int) -> None:
    r"""Set environment variables used to limit number of process threads."""
    if threads < 0:
        raise ValueError("init_num_threads() 'threads' must not be negative")
    threads = max(1, int(threads))
    for env in (
        "ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS",
        "MKL_NUM_THREADS",
        "NUMEXPR_NUM_THREADS",
        "OMP_NUM_THREADS",
    ):
        os.environ[env] = str(threads)
    try:
        import SimpleITK as sitk

        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(threads)
    except ImportError:
        pass


def init_omp_num_threads(threads: Optional[int] = None, default: int = 1) -> int:
    r"""Set environment variable OMP_NUM_THREADS to limit number of process threads."""
    if threads is None or threads < 0:
        threads = os.environ.get("OMP_NUM_THREADS", default)
    threads = max(1, int(threads))
    os.environ["OMP_NUM_THREADS"] = str(threads)
    return threads



r"""Test basic network layers."""

import pytest

import torch
from torch import Tensor, nn

from deepali.core import PaddingMode
from deepali.networks.layers import Activation, activation
from deepali.networks.layers import Conv2d, Conv3d, ConvLayer, convolution, conv_module
from deepali.networks.layers import JoinLayer, join_func
from deepali.networks.layers import LambdaLayer
from deepali.networks.layers import NormLayer, normalization, norm_layer


def test_activation() -> None:
    r"""Test construction of non-linear activations."""

    act = activation("relu")
    assert isinstance(act, nn.ReLU)
    act = activation("lrelu", 0.1)
    assert isinstance(act, nn.LeakyReLU)
    assert act.negative_slope == 0.1
    assert act.inplace is False
    act = activation(("LeakyReLU", {"negative_slope": 0.05}), inplace=True)
    assert isinstance(act, nn.LeakyReLU)
    assert act.negative_slope == 0.05
    assert act.inplace is True
    act = activation("softmax", dim=2)
    assert isinstance(act, nn.Softmax)
    assert act.dim == 2

    def act_func(x: Tensor) -> Tensor:
        return x.sigmoid()

    act = activation(act_func)
    assert isinstance(act, Activation)
    assert isinstance(act, nn.Module)
    assert act.func is act_func

    x = torch.tensor(2.0)
    y = act(x)
    assert isinstance(y, Tensor)
    assert y.allclose(act_func(x))

    act = Activation("elu")
    assert isinstance(act.func, nn.ELU)


def test_convolution() -> None:
    r"""Test construction of simple convolutional layer."""

    conv = convolution(1, 3, 16, 1)
    assert isinstance(conv, nn.Conv1d)
    assert conv.weight.shape == (16, 3, 1)
    assert isinstance(conv.bias, Tensor)

    conv = convolution(2, 3, 32, kernel_size=1)
    assert isinstance(conv, nn.Conv2d)
    assert isinstance(conv, Conv2d)
    assert conv.weight.shape == (32, 3, 1, 1)
    assert isinstance(conv.bias, Tensor)

    conv = convolution(3, 3, 8, kernel_size=1)
    assert isinstance(conv, nn.Conv3d)
    assert isinstance(conv, Conv3d)
    assert conv.weight.shape == (8, 3, 1, 1, 1)
    assert isinstance(conv.bias, Tensor)

    conv = convolution(3, 3, 8, 5, bias=False)
    assert isinstance(conv, nn.Conv3d)
    assert conv.weight.shape == (8, 3, 5, 5, 5)
    assert conv.bias is None
    assert conv.output_padding == (0, 0, 0)

    conv = convolution(3, 3, 8, kernel_size=(5, 3, 1), output_padding=0)
    assert isinstance(conv, nn.Conv3d)
    assert conv.weight.shape == (8, 3, 5, 3, 1)
    assert isinstance(conv.bias, Tensor)
    assert conv.output_padding == (0, 0, 0)

    conv = conv_module(3, 3, 16, kernel_size=3, stride=2, output_padding=1, transposed=True)
    assert isinstance(conv, nn.ConvTranspose3d)
    assert conv.transposed is True
    assert conv.kernel_size == (3, 3, 3)
    assert conv.stride == (2, 2, 2)
    assert conv.output_padding == (1, 1, 1)

    conv = conv_module(
        spatial_dims=3,
        in_channels=6,
        out_channels=16,
        kernel_size=3,
        stride=2,
        dilation=4,
        padding=1,
        padding_mode=PaddingMode.REFLECT,
        groups=2,
        init="xavier",
        bias="zeros",
    )
    assert isinstance(conv, Conv3d)
    assert conv.weight_init == "xavier"
    assert conv.bias_init == "zeros"
    assert conv.in_channels == 6
    assert conv.out_channels == 16
    assert conv.groups == 2
    assert conv.padding_mode == "reflect"
    assert conv.kernel_size == (3, 3, 3)
    assert conv.stride == (2, 2, 2)
    assert conv.dilation == (4, 4, 4)
    assert conv.padding == (1, 1, 1)
    assert conv.output_padding == (0, 0, 0)
    assert conv.transposed is False


def test_conv_layer() -> None:
    r"""Test convolutional layer with optional normalization and/or activation."""
    layer = ConvLayer(2, 1, 8, 3)
    assert isinstance(layer, nn.Module)
    assert hasattr(layer, "acti")
    assert hasattr(layer, "norm")
    assert hasattr(layer, "conv")
    assert layer.acti is None
    assert layer.norm is None
    assert isinstance(layer.conv, nn.Conv2d)

    layer = ConvLayer(
        1, in_channels=1, out_channels=16, kernel_size=3, acti=("lrelu", {"negative_slope": 0.1})
    )
    assert isinstance(layer.acti, nn.LeakyReLU)
    assert layer.norm is None
    assert isinstance(layer.conv, nn.Conv1d)
    assert layer.order == "CNA"
    assert layer.acti.negative_slope == 0.1

    layer = ConvLayer(2, in_channels=1, out_channels=16, kernel_size=3, acti="relu")
    assert isinstance(layer.acti, nn.ReLU)
    assert layer.norm is None
    assert isinstance(layer.conv, nn.Conv2d)
    assert layer.order == "CNA"

    layer = ConvLayer(3, in_channels=1, out_channels=16, kernel_size=3, acti="prelu", norm="batch")
    assert isinstance(layer.acti, nn.PReLU)
    assert isinstance(layer.norm, nn.BatchNorm3d)
    assert isinstance(layer.conv, nn.Conv3d)
    assert layer.order == "CNA"

    layer = ConvLayer(
        3, in_channels=1, out_channels=16, kernel_size=3, acti="relu", norm="instance", order="nac"
    )
    assert isinstance(layer.acti, nn.ReLU)
    assert isinstance(layer.norm, nn.InstanceNorm3d)
    assert isinstance(layer.conv, nn.Conv3d)
    assert layer.order == "NAC"

    x = torch.randn((2, 1, 7, 9, 11))
    y = layer(x)
    z = layer.conv(layer.acti(layer.norm(x)))
    assert isinstance(y, Tensor)
    assert isinstance(z, Tensor)
    assert y.allclose(z)

    layer = ConvLayer(
        3, in_channels=1, out_channels=16, kernel_size=3, acti="relu", norm="instance", order="CNA"
    )
    assert isinstance(layer.acti, nn.ReLU)
    assert isinstance(layer.norm, nn.InstanceNorm3d)
    assert isinstance(layer.conv, nn.Conv3d)
    assert layer.order == "CNA"

    y = layer(x)
    assert isinstance(y, Tensor)
    assert not y.allclose(z)
    z = layer.acti(layer.norm(layer.conv(x)))
    assert isinstance(z, Tensor)
    assert y.allclose(z)


def test_join_layer() -> None:
    r"""Test layer which joins features of one or more input tensors."""

    with pytest.raises(ValueError):
        join_func("foo")
    with pytest.raises(ValueError):
        JoinLayer("bar")

    x = torch.tensor([[1.0, 1.5, 2.0]])
    y = torch.tensor([[0.4, 2.5, -0.1]])

    func = join_func("add")
    z = func([x, y])
    assert isinstance(z, Tensor)
    assert z.allclose(x + y)

    func = join_func("mul")
    z = func([x, y])
    assert isinstance(z, Tensor)
    assert z.allclose(x * y)

    func = join_func("cat")
    z = func([x, y])
    assert isinstance(z, Tensor)
    assert z.allclose(torch.cat([x, y], dim=1))

    func = join_func("concat")
    z = func([x, y])
    assert isinstance(z, Tensor)
    assert z.allclose(torch.cat([x, y], dim=1))

    join = JoinLayer("add", dim=0)
    z = join([x, y])
    assert isinstance(z, Tensor)
    assert z.allclose(x + y)

    join = JoinLayer("mul")
    z = join([x, y])
    assert isinstance(z, Tensor)
    assert z.allclose(x * y)

    join = JoinLayer("cat", dim=0)
    z = join([x, y])
    assert isinstance(z, Tensor)
    assert z.allclose(torch.cat([x, y], dim=0))

    join = JoinLayer("concat", dim=1)
    z = join([x, y])
    assert isinstance(z, Tensor)
    assert z.allclose(torch.cat([x, y], dim=1))


def test_lambda_layer() -> None:
    def square_func(x: Tensor) -> Tensor:
        return x.square()

    square_layer = LambdaLayer(square_func)
    assert isinstance(square_layer, nn.Module)
    assert square_layer.func is square_func

    x = torch.tensor([[2.0, 0.5]])
    y = square_layer(x)
    assert isinstance(y, Tensor)
    assert y.allclose(torch.tensor([[4.0, 0.25]]))


def test_norm_layer() -> None:
    r"""Test construction of normalization layers."""

    # Batch normalization
    with pytest.raises(ValueError):
        norm_layer("batch", spatial_dims=3)
    with pytest.raises(ValueError):
        normalization("batch", spatial_dims=1)
    with pytest.raises(ValueError):
        normalization("batch", num_features=32)
    norm = norm_layer("batch", spatial_dims=1, num_features=16)
    assert isinstance(norm, nn.BatchNorm1d)
    assert norm.affine is True
    assert norm.num_features == 16
    assert norm.bias.shape == (16,)
    norm = normalization("batch", spatial_dims=2, num_features=32)
    assert isinstance(norm, nn.BatchNorm2d)
    assert norm.affine is True
    assert norm.num_features == 32
    assert norm.bias.shape == (32,)
    norm = norm_layer("BatchNorm", spatial_dims=3, num_features=64)
    assert isinstance(norm, nn.BatchNorm3d)
    assert norm.affine is True
    assert norm.num_features == 64
    assert norm.bias.shape == (64,)

    # Group normalization
    with pytest.raises(ValueError):
        norm_layer("group", 3)
    with pytest.raises(ValueError):
        norm_layer("group", spatial_dims=1)
    norm = norm_layer("group", num_features=32)
    assert isinstance(norm, nn.GroupNorm)
    assert norm.num_groups == 1
    assert norm.num_channels == 32
    assert norm.affine is True
    norm = norm_layer("GroupNorm", num_channels=16)
    assert isinstance(norm, nn.GroupNorm)
    assert norm.num_groups == 1
    assert norm.num_channels == 16
    assert norm.affine is True
    norm = normalization("group", num_groups=8, num_channels=64, affine=False)
    assert isinstance(norm, nn.GroupNorm)
    assert norm.num_groups == 8
    assert norm.num_channels == 64
    assert norm.affine is False

    # Layer normalization
    with pytest.raises(ValueError):
        norm_layer("layer")
    with pytest.raises(ValueError):
        norm_layer("layer", spatial_dims=1)
    norm = normalization("layer", num_features=32)
    assert isinstance(norm, nn.GroupNorm)
    assert norm.num_groups == 1
    assert norm.num_channels == 32
    assert norm.affine is True
    norm = norm_layer("LayerNorm", spatial_dims=1, num_features=64, affine=False)
    assert isinstance(norm, nn.GroupNorm)
    assert norm.num_groups == 1
    assert norm.num_channels == 64
    assert norm.affine is False

    # Instance normalization
    with pytest.raises(ValueError):
        norm_layer("instance", 2)
    with pytest.raises(ValueError):
        norm_layer("instance", spatial_dims=1)
    with pytest.raises(ValueError):
        normalization("instance", num_features=32)
    norm = norm_layer("instance", spatial_dims=1, num_features=32)
    assert isinstance(norm, nn.InstanceNorm1d)
    norm = norm_layer("instance", spatial_dims=2, num_features=32)
    assert isinstance(norm, nn.InstanceNorm2d)
    norm = normalization("instance", spatial_dims=3, num_features=32)
    assert isinstance(norm, nn.InstanceNorm3d)
    assert norm.num_features == 32
    assert norm.affine is False
    norm = normalization("InstanceNorm", spatial_dims=3, num_features=32, affine=True)
    assert isinstance(norm, nn.InstanceNorm3d)
    assert norm.num_features == 32
    assert norm.affine is True
    norm = NormLayer("instance", spatial_dims=3, num_features=64)
    assert isinstance(norm.func, nn.InstanceNorm3d)
    assert norm.func.num_features == 64
    assert norm.func.affine is False

    # Custom normalization
    def norm_func(x: Tensor) -> Tensor:
        return x

    norm = normalization(norm_func)
    assert isinstance(norm, nn.Module)
    assert isinstance(norm, NormLayer)
    assert norm.func is norm_func

    norm = NormLayer(norm_func)
    assert norm.func is norm_func



r"""File path utility functions."""

from errno import ENOTEMPTY
import os
from os import rmdir
from pathlib import Path
import re
from shutil import rmtree
from typing import Any, Optional, Union, overload
from urllib.parse import urlsplit

from .typing import PathStr, PathUri, is_path_str_field, is_path_str, is_path_str_type_hint


__all__ = (
    "Path",
    "PathStr",
    "PathUri",
    "abspath",
    "abspath_template",
    "delete",
    "is_absolute",
    "is_path_str",
    "is_path_str_field",
    "is_path_str_type_hint",
    "is_uri",
    "make_parent_dir",
    "norm_uri",
    "path_suffix",
    "path_uri",
    "to_uri",
    "unlink_or_mkdir",
)


@overload
def abspath(path: str, parent: Optional[PathStr] = None) -> str:
    ...


@overload
def abspath(path: Path, parent: Optional[PathStr] = None) -> Path:
    ...


def abspath(path: PathStr, parent: Optional[PathStr] = None) -> PathStr:
    r"""Make path absolute."""
    path_is_str = isinstance(path, str)
    if not path_is_str and not isinstance(path, Path):
        raise TypeError(f"abspath() 'path' must be pathlib.Path or str, not {type(path)}")
    path = (Path(parent or ".") / Path(path)).absolute()
    if path_is_str:
        path = path.as_posix()
    return path


@overload
def abspath_template(path: str, parent: Optional[Path] = None) -> str:
    ...


@overload
def abspath_template(path: Path, parent: Optional[Path] = None) -> Path:
    ...


def abspath_template(path: PathStr, parent: Optional[Path] = None) -> PathStr:
    r"""Make path format string absolute."""
    if not isinstance(path, (Path, str)):
        raise TypeError(f"abspath_template() 'path' must be pathlib.Path or str, not {type(path)}")
    if str(path).startswith("{"):
        return path
    return abspath(path, parent=parent)


def delete(path: PathStr, non_empty: bool = True) -> bool:
    r"""Remove file or (non-empty) directory."""
    path = Path(path)
    if path.is_dir():
        if non_empty:
            rmtree(path)
        else:
            try:
                rmdir(path)
            except OSError as error:
                if error.errno == ENOTEMPTY:
                    return False
                raise
    else:
        try:
            path.unlink()
        except FileNotFoundError:
            return False
    return True


def is_absolute(path: Union[Path, str]) -> bool:
    r"""Check whether given path string or URI is absolute."""
    if is_uri(path):
        return True
    return Path(path).is_absolute()


def is_uri(arg: Any) -> bool:
    r"""Check whether a given argument is a URI."""
    if isinstance(arg, Path):
        return False
    if isinstance(arg, str):
        # Windows path with drive letter
        if os.name == "nt" and re.match(r"([a-zA-Z]):[/\\](.*)", arg):
            return False
        return re.match(r"([a-zA-Z0-9]+)://(.*)", arg) is not None
    return False


def make_parent_dir(path: PathStr, parents: bool = True, exist_ok: bool = True) -> Path:
    r"""Make parent directory of file path."""
    parent = Path(path).absolute().parent
    parent.mkdir(parents=parents, exist_ok=exist_ok)
    return parent


def norm_uri(uri: str) -> str:
    r"""Normalize URI.

    Args:
        uri: A valid URI string.

    Returns:
        Normalized URI string.

    """
    match = re.match(r"(?P<scheme>[a-zA-Z0-9]+)://(?P<path>.*)", uri)
    if not match:
        raise ValueError(f"norm_uri() 'uri' is not a valid URI: {uri}")
    scheme = match["scheme"].lower()
    path = re.sub("^/+", "", re.sub(r"[/\\]{1,}", "/", match["path"]))
    # Local file URI
    if scheme == "file":
        if os.name != "nt" or re.match(r"(?P<drive>[a-zA-Z]):[/\\]", path) is None:
            path = "/" + path
        return "file://" + path
    # AWS S3 object URI
    if scheme == "s3":
        return "s3://" + path
    # Other URI
    return urlsplit(uri, scheme="file").geturl()


def path_suffix(path: PathUri) -> str:
    r"""Get filename suffix, including .gz, .bz2 if present."""
    m = re.search(r"(\.[a-zA-Z0-9]+(\.gz|\.GZ|\.bz2|\.BZ2|\.zst|\.ZST)?)$", str(path))
    return m.group(1) if m else ""


def path_uri(arg: PathStr) -> str:
    r"""Create valid URI from local path.

    Unlike Path.as_uri(), this function does not escape special characters as used in format template strings.

    Args:
        arg: Local path.

    Returns:
        Valid URI.

    """
    uri = norm_uri(f"file://{Path(arg).absolute()}")
    if uri.endswith("/"):
        # Trailing forward slashes are removed by Path already, but trailing backward slashes are
        # only converted to a single forward slash by norm_uri(), not removed by Path. To produce
        # a consistent result regardless of whether forward or backward slashes are used, remove
        # any remaining trailing slash even if it could signify a directory.
        uri = uri[:-1]
    return uri


def to_uri(*args: Optional[PathUri]) -> str:
    r"""Create valid URI from resource paths.

    Args:
        args: Local path components or an already valid URI. The last absolute path or URI in this
            list of arguments is the base path or URI prefix for subsequent relative paths which
            are appended to this base to construct the URI. Any ``None`` values are ignored.

    Returns:
        Valid URI.

    """
    args = [arg for arg in args if arg is not None]
    for i, arg in enumerate(reversed(args)):
        if is_uri(arg):
            base = str(arg)
            args = args[len(args) - i :]
            break
        elif Path(arg).is_absolute():
            base = Path(arg)
            args = args[len(args) - i :]
            break
    else:
        base = Path.cwd()
    if isinstance(base, Path):
        uri = path_uri(base.joinpath(*args))
    else:
        uri = norm_uri(f"{base}/{'/'.join(args)}" if args else base)
    return uri


def unlink_or_mkdir(path: PathStr) -> Path:
    r"""Unlink existing file or make parent directory if non-existent.

    This function is useful when a script output file is managed by DVC using protected symbolic or hard links.
    Call this function before writing the new output file. It will unlike any existing output file, and ensure
    that the output directory exists.

    Args:
        path: File path.

    Returns:
        Absolute file path.

    """
    path = Path(path).absolute()
    try:
        path.unlink()
    except FileNotFoundError:
        path.parent.mkdir(parents=True, exist_ok=True)
    return path



r"""Attributes of data sampling grid oriented in space."""

import itertools
from typing import Sequence, Tuple, TypeVar, Union, Optional

import numpy as np
from numpy.typing import ArrayLike, NDArray
import SimpleITK as sitk


TGridAttrs = TypeVar("TGridAttrs", bound="GridAttrs")


class GridAttrs(object):
    r"""Attributes of a regular data sampling grid oriented in world space."""

    __slots__ = ["_size", "origin", "spacing", "direction"]

    def __init__(
        self,
        size: Sequence[Union[float, int]],
        center: Optional[Sequence[float]] = None,
        origin: Optional[Sequence[float]] = None,
        spacing: Optional[Sequence[float]] = None,
        direction: Optional[Union[Sequence[float], Sequence[Sequence[float]]]] = None,
    ) -> None:
        r"""Initialize grid attributes.

        Note that the order of spatial dimensions of the attributes is ``(X, ...)``,
        whereas the data arrays often are given in reverse order, i.e., the shape of a
        corresponding data array is ``(..., X)``.

        Args:
            size: Size of spatial grid dimensions. The length determines the number of dimensions.
            center: Grid center point coordinates in world space along each dimension.
            origin: World coordinates of grid point with index zero along each dimension.
            spacing: Spacing between grid points along each dimension.
            direction: Direction cosines. Must be a square matrix, but can be flattened.

        """
        ndim = len(size)
        if spacing is None:
            spacing = (1.0,) * ndim
        elif not isinstance(spacing, tuple):
            spacing = tuple(spacing)
        if direction is None:
            direction = np.eye(ndim, ndim)
        direction = tuple(float(x) for x in np.asanyarray(direction).flatten())
        if origin is None:
            if center is None:
                origin = (0.0,) * ndim
            else:
                offset = np.array(0.5 * n if n > 0 else 0 for n in self._int_size(size))
                rotation = np.array(self.direction).reshape(ndim, ndim)
                scaling = np.diag(spacing)
                coords: NDArray = np.asanyarray(center) - np.matmul(rotation @ scaling, offset)
                origin = tuple(float(x) for x in coords)
        elif center is not None:
            raise ValueError("Grid() 'center' and 'origin' are mutually exclusive")
        elif not isinstance(origin, tuple):
            origin = tuple(origin)
        # store grid size as float such that grid.down().up() restores original grid
        self._size = tuple(float(n) for n in size)
        self.origin = origin
        self.spacing = spacing
        self.direction = direction

    @property
    def center(self) -> Tuple[float, ...]:
        r"""Get grid center point coordinates in world space."""
        offset = np.array(0.5 * n if n > 0 else 0 for n in self.size)
        coords: NDArray = self.origin + np.matmul(self.transform[:-1, :-1], offset)
        return tuple(float(x) for x in coords)

    @property
    def ndim(self) -> int:
        r"""Number of spatial grid dimensions."""
        return len(self.shape)

    @property
    def npoints(self) -> int:
        r"""Total number of grid points."""
        return int(np.prod(self.size))

    @staticmethod
    def _int_size(size) -> Tuple[int, ...]:
        """Get grid size from internal floating point representation."""
        return tuple([int(n + 0.5) if n > 1 or n <= 0 else 1 for n in size])

    @property
    def size(self) -> Tuple[int, ...]:
        r"""Size of sampling grid with spatial dimensions in the order (X, ...)."""
        return self._int_size(self._size)

    @property
    def shape(self) -> Tuple[int, ...]:
        r"""Shape of sampling grid with spatial dimensions in the order (..., X)."""
        return tuple(reversed(self.size))

    def with_margin(self: TGridAttrs, margin: Union[int, None]) -> TGridAttrs:
        r"""Create new image grid with an additional margin along each grid axis."""
        if not margin:
            return self
        cls = type(self)
        return cls(
            size=tuple(n + 2 * margin for n in self._size),
            origin=self.index_to_physical_space((-margin,) * self.ndim),
            spacing=self.spacing,
            direction=self.direction,
        )

    def with_spacing(self: TGridAttrs, *args: float) -> TGridAttrs:
        r"""Create new image grid with specified spacing."""
        cls = type(self)
        spacing = np.asarray(*args)
        assert 0 <= spacing.ndim <= 1
        if spacing.ndim == 0:
            spacing = spacing.repeat(self.ndim)
        assert spacing.size == self.ndim
        cur_size = np.asarray(self._size)
        cur_spacing = np.asarray(self.spacing)
        size = cur_size * cur_spacing / spacing
        shift = 0.5 * (np.round(cur_size) - np.round(size) * spacing / cur_spacing)
        origin = self.index_to_physical_space(shift)
        return cls(size=size, origin=origin, spacing=spacing, direction=self.direction)

    def down(self: TGridAttrs, levels: int = 1) -> TGridAttrs:
        r"""Create new sampling grid of half the size while preserving grid extent."""
        cls = type(self)
        size = self._size
        for _ in range(levels):
            size = tuple(n / 2 for n in size)
        cur_size = self.size
        new_size = self._int_size(size)
        spacing = tuple(
            self.spacing[i] * cur_size[i] / new_size[i] if new_size[i] > 0 else self.spacing[i]
            for i in range(self.ndim)
        )
        return cls(size=size, center=self.center, spacing=spacing, direction=self.direction)

    def up(self: TGridAttrs, levels: int = 1) -> TGridAttrs:
        r"""Create new sampling grid of double the size while preserving grid extent."""
        cls = type(self)
        size = self._size
        for _ in range(levels):
            size = tuple(2 * n for n in size)
        cur_size = self.size
        new_size = self._int_size(size)
        spacing = tuple(
            self.spacing[i] * cur_size[i] / new_size[i] if new_size[i] > 0 else self.spacing[i]
            for i in range(self.ndim)
        )
        return cls(size=size, center=self.center, spacing=spacing, direction=self.direction)

    @property
    def dcm(self) -> np.ndarray:
        r"""Get direction cosine matrix."""
        return np.array(self.direction).reshape(self.ndim, self.ndim)

    @property
    def transform(self) -> np.ndarray:
        r"""Get homogeneous coordinate transformation from image grid to world space."""
        rotation = self.dcm
        scaling = np.diag(self.spacing)
        matrix = homogeneous_matrix(rotation @ scaling)
        matrix[0:-1, -1] = self.origin
        return matrix

    @property
    def inverse_transform(self) -> np.ndarray:
        r"""Get homogeneous coordinate transformation from world space to image grid."""
        rotation = self.dcm.T
        scaling = np.diag([1 / s for s in self.spacing])
        translation = translation_matrix([-t for t in self.origin])
        return homogeneous_matrix(scaling @ rotation) @ translation

    @property
    def indices(self) -> np.ndarray:
        r"""Get array of image grid point coordinates in image space."""
        return np.flip(
            np.stack(
                np.meshgrid(*[np.arange(arg) for arg in self.shape], indexing="ij"),
                axis=self.ndim,
            ),
            axis=-1,
        )

    def axis_indices(self, axis: int) -> np.ndarray:
        r"""Get array of image grid indices along specified axis."""
        return np.arange(self.shape[axis])

    @property
    def points(self) -> np.ndarray:
        r"""Get array of image grid point coordinates in world space."""
        return self.index_to_physical_space(self.indices)

    @property
    def coords(self) -> Tuple[np.ndarray, ...]:
        r"""Get 1D arrays of grid point coordinates along each axis."""
        return tuple(self.axis_coords(axis) for axis in range(self.ndim))

    def axis_coords(self, axis: int) -> np.ndarray:
        r"""Get array of image grid point coordinates in world space along specified axis."""
        indices = [[0]] * self.ndim
        indices[axis] = [index for index in range(self.shape[axis])]
        mesh = np.stack(np.meshgrid(*indices, indexing="ij"), axis=self.ndim)
        return self.index_to_physical_space(mesh)[..., axis].flatten()

    @property
    def corners(self) -> np.ndarray:
        r"""Get corners of image domain in world space."""
        limits = []
        for axis in range(self.ndim):
            limits.append((0, self.shape[axis]))
        corners = [tuple(reversed(indices)) for indices in itertools.product(*limits)]
        return self.index_to_physical_space(corners)

    def index_to_physical_space(self, points: ArrayLike) -> np.ndarray:
        r"""Map point coordinates from image to world space.

        Args:
            points: Input voxel indices as n-dimensional array of shape ``(..., D)``.
                The input can be both integral or floating point indices.

        Returns:
            World coordinates of the indexed grid locations.

        """
        return transform_point(self.transform, points)

    def physical_space_to_index(self, points: ArrayLike) -> np.ndarray:
        r"""Map point coordinates from world to discrete image space.

        Args:
            points: World coordinates of input points given as n-dimensional array of shape ``(..., D)``.

        Returns:
            Integral indices of nearest grid points.

        """
        index: np.ndarray = np.round(self.physical_space_to_continuous_index(points))
        return index.astype(int)

    def physical_space_to_continuous_index(self, points: ArrayLike) -> np.ndarray:
        r"""Map point coordinates from world to continuous image space.

        Args:
            points: World coordinates of input points given as n-dimensional array of shape ``(..., D)``.

        Returns:
            Indices of grid locations as floating point values.

        """
        # Round to avoid 1e-15, 1e-17, -1e-14,... as representations for zero, s.t.,
        # indices == physical_space_to_continuous_index(index_to_physical_space(indices))
        return np.round(transform_point(self.inverse_transform, points), decimals=12)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"size={self.size}"
            f", origin={self.origin}"
            f", spacing={self.spacing}"
            f", direction={self.direction}"
            ")"
        )


def image_grid_attributes(image: sitk.Image) -> GridAttrs:
    r"""Get image sampling grid attributes."""
    return GridAttrs(
        size=image.GetSize(),
        origin=image.GetOrigin(),
        spacing=image.GetSpacing(),
        direction=image.GetDirection(),
    )


def homogeneous_coords(point: np.ndarray, ndim: int = None, copy: bool = True) -> np.ndarray:
    r"""Create array of homogeneous point coordinates from D-dimensional point set.

    Args:
        point: Input point or point set as array of shape ``(..., D)``.
        ndim: Expected number of spatial dimensions. If specified, and the input ``point``
            array is of shape ``(..., ndim + 1)``, the input point array is returned
            without adding an additional dimension.
        copy: Whether to copy the input ``point`` array when ``ndim`` is specified and
            the shape of the input array is ``(..., ndim + 1)``.

    Returns:
        Array of homogeneous coordinates with shape ``(..., D + 1)``.

    """
    if ndim is not None:
        if point.shape[-1] == ndim + 1:
            return np.copy(point) if copy else point
        assert point.shape[-1] == ndim
    # Creating new array with submatrix assignment is faster than np.hstack!
    pts = np.ones(point.shape[0:-1] + (point.shape[-1] + 1,), dtype=point.dtype)
    pts[..., :-1] = point
    return pts


def homogeneous_matrix(transform: np.ndarray, ndim: int = None, copy: bool = True) -> np.ndarray:
    r"""Create homogeneous transformation matrix from affine coordinate transformation."""
    assert transform.ndim == 2
    rows, cols = transform.shape
    if ndim is None:
        ndim = rows
        assert (
            transform.shape[1] == ndim or transform.shape[1] == ndim + 1
        ), "transform.shape={}".format(transform.shape)
    elif rows == ndim + 1 and cols == ndim + 1:
        return np.copy(transform) if copy else transform
    matrix = np.eye(ndim + 1, ndim + 1, dtype=transform.dtype)
    matrix[0:rows, 0:cols] = transform
    return matrix


def translation_matrix(displacement: ArrayLike) -> np.ndarray:
    r"""Create translation matrix for homogeneous coordinates."""
    arg = np.asanyarray(displacement)
    assert arg.ndim == 1
    matrix = np.eye(arg.size + 1)
    matrix[0:-1, -1] = arg
    return matrix


def transform_point(matrix: np.ndarray, point: ArrayLike) -> np.ndarray:
    r"""Transform one or more points given a transformation matrix."""
    arg = np.asanyarray(point)
    dim = arg.shape[-1]
    mat = homogeneous_matrix(matrix, ndim=dim, copy=False)
    x = arg.reshape(-1, dim)
    y: np.ndarray = np.matmul(x, mat[0:-1, 0:-1].T) + np.expand_dims(mat[0:-1, -1], axis=0)
    return y.reshape(arg.shape)


def transform_vector(matrix: np.ndarray, vector: ArrayLike) -> np.ndarray:
    r"""Transform one or more vectors given a transformation matrix."""
    arg = np.asanyarray(vector)
    dim = arg.shape[-1]
    v = arg.reshape(-1, dim)
    u: np.ndarray = np.matmul(v, matrix[0:dim, 0:dim].T)
    return u.reshape(arg.shape)



r"""Convolutional network with residual shortcut connections.

For further reference implementations of 2D and 3D ResNet, see for example:

- https://github.com/pytorch/vision/blob/1703e4ca4f879f509e797ea816670ee72fae55dd/torchvision/models/resnet.py
- https://github.com/kenshohara/3D-ResNets-PyTorch/blob/540a0ea1abaee379fa3651d4d5afbd2d667a1f49/models/resnet.py

"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Mapping, Optional, Sequence, Type, Union

from torch.nn import Flatten, Module, Sequential, init

from deepali.core.config import DataclassConfig
from deepali.core.enum import PaddingMode
from deepali.core.itertools import zip_longest_repeat_last
from deepali.core.typing import ScalarOrTuple
from deepali.modules import ReprWithCrossReferences

from .blocks import ResidualUnit
from .layers import ActivationArg, NormArg, ConvLayer, Linear, Upsample, pooling
from .layers import is_batch_norm, is_convolution, is_group_norm


# fmt: off
__all__ = (
    "ResidualUnit",
    "ResNet",
    "ResNetConfig",
)
# fmt: on


ModuleFactory = Union[Callable[..., Module], Type[Module]]


MODEL_DEPTHS = {10, 18, 34, 50, 101, 152, 200}


@dataclass
class ResNetConfig(DataclassConfig):
    r"""Configuration of residual network architecture."""

    # Input shape
    spatial_dims: int
    in_channels: int = 1

    # cf. ResNet and ResidualUnit
    stride: ScalarOrTuple[int] = (1, 2, 2, 2)
    num_blocks: ScalarOrTuple[int] = (3, 4, 6, 3)
    num_channels: Sequence[int] = (64, 128, 256, 512)
    num_layers: int = 2
    num_classes: Optional[int] = None
    kernel_size: int = 3
    expansion: float = 1
    padding_mode: Union[PaddingMode, str] = "zeros"
    pre_conv: bool = False
    post_deconv: bool = False
    recursive: ScalarOrTuple[bool] = False
    bias: bool = False
    norm: NormArg = "batch"
    acti: ActivationArg = "relu"
    order: str = "cna"
    skip: str = "identity | conv1"
    residual_pre_conv: str = "conv1"
    residual_post_conv: str = "conv1"

    @classmethod
    def from_depth(
        cls, model_depth: int, spatial_dims: int, in_channels: int = 1, **kwargs
    ) -> ResNetConfig:
        r"""Get default ResNet configuration for given depth."""
        config = cls(spatial_dims=spatial_dims, in_channels=in_channels, **kwargs)
        if model_depth == 10:
            config.num_blocks = (1, 1, 1, 1)
            config.num_layers = 2
            config.expansion = 1
        elif model_depth == 18:
            config.num_blocks = (2, 2, 2, 2)
            config.num_layers = 2
            config.expansion = 1
        elif model_depth == 34:
            config.num_blocks = (3, 4, 6, 3)
            config.num_layers = 2
            config.expansion = 1
        elif model_depth == 50:
            config.num_blocks = (3, 4, 6, 3)
            config.num_layers = 3
            config.expansion = 4
        elif model_depth == 101:
            config.num_blocks = (3, 4, 23, 3)
            config.num_layers = 3
            config.expansion = 4
        elif model_depth == 152:
            config.num_blocks = (3, 8, 36, 3)
            config.num_layers = 3
            config.expansion = 4
        elif model_depth == 200:
            config.num_blocks = (3, 24, 36, 3)
            config.num_layers = 3
            config.expansion = 4
        else:
            raise ValueError("ResNetConfig.from_depth() 'model_depth' must be in {MODEL_DEPTHS!r}")
        return config


def classification_head(spatial_dims: int, in_channels: int, num_classes: int, **kwargs) -> Module:
    r"""Image classification head for ResNet model."""
    pool = pooling("AdaptiveAvg", spatial_dims=spatial_dims, output_size=1)
    fc = Linear(in_channels, num_classes)
    return Sequential(pool, Flatten(), fc)


def conv_layer(
    level: int,
    is_first: bool,
    spatial_dims: int,
    in_channels: int,
    out_channels: int,
    **kwargs,
) -> Module:
    r"""Convolutional layer before/between residual blocks when ``pre_conv=True``."""
    norm = kwargs.get("norm")
    acti = kwargs.get("acti")
    order = kwargs.get("order", "cna").upper()
    if is_first:
        if "N" in order and order.index("N") < order.index("C"):
            norm = None
        if "A" in order and order.index("A") < order.index("C"):
            acti = None
    kwargs.update(dict(norm=norm, acti=acti, order=order))
    return ConvLayer(spatial_dims, in_channels, out_channels, **kwargs)


def input_layer(spatial_dims: int, in_channels: int, out_channels: int, **kwargs) -> Module:
    r"""First convolutional ResNet layer."""
    order = kwargs.get("order", "cna").upper()
    norm = None if "N" in order and order.index("N") < order.index("C") else kwargs.get("norm")
    kwargs.update(dict(kernel_size=7, padding=3, stride=2, norm=norm, acti=None, order=order))
    conv = ConvLayer(spatial_dims, in_channels, out_channels, **kwargs)
    pool = pooling("max", spatial_dims=spatial_dims, kernel_size=3, stride=2, padding=1)
    return Sequential(conv, pool)


class ResNet(ReprWithCrossReferences, Sequential):
    r"""Residual network.

    Note that unlike ``torchvision.models.ResNet``, the ``__init__`` function of this class
    does not initialize the parameters of the model, other than the standard initialization
    for each module type. In order to apply the initialization of the torchvision ResNet, call
    functions ``init_conv_modules()``, ``init_norm_layers()``, and ``zero_init_residuals()``
    (in this order!) after constructing the ResNet model.

    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int = 1,
        num_channels: ScalarOrTuple[int] = (64, 128, 256, 512),
        num_blocks: ScalarOrTuple[int] = (3, 4, 6, 3),
        num_layers: int = 2,
        num_classes: Optional[int] = None,
        kernel_size: int = 3,
        stride: ScalarOrTuple[int] = (1, 2, 2, 2),
        expansion: float = 1,
        padding_mode: Union[PaddingMode, str] = "zeros",
        recursive: ScalarOrTuple[bool] = False,
        pre_conv: bool = False,
        post_deconv: bool = False,
        bias: Union[bool, str] = False,
        norm: NormArg = "batch",
        acti: ActivationArg = "relu",
        order: str = "cna",
        skip: str = "identity | conv1",
        residual_pre_conv: str = "conv1",
        residual_post_conv: str = "conv1",
        conv_layer: ModuleFactory = conv_layer,
        deconv_layer: ModuleFactory = Upsample,
        resnet_block: ModuleFactory = ResidualUnit,
        input_layer: Optional[ModuleFactory] = input_layer,
        output_layer: Optional[ModuleFactory] = None,
    ) -> None:
        r"""Initialize layers.

        Args:
            spatial_dims: Number of spatial tensor dimensions.
            in_channels: Number of input channels.
            num_channels: Number of feature channels at each level.
            num_blocks: Number of residual blocks at each level.
            num_layers: Number of convolutional layers in each residual block.
            num_classes: Number of output class probabilities of ``output_layer``.
            kernel_size: Size of convolutional filters in residual blocks.
            stride: Stride of initial convolution at each level. Subsequent convolutions have stride 1.
            expansion: Expansion factor of ``num_channels``. If specified, the number of input and output
                feature maps for each residual block are equal to ``expansion * num_channels``, and the
                bottleneck convolutional layers after an initial convolution with kernel size 1 operate
                on feature maps with ``num_channels`` each, which are subsequently expanded again by the
                specified expansion level by another convolution with kernel size 1.
            padding_mode: Padding mode for convolutional layers with kernel size greater than 1.
            recursive: Whether residual blocks at each level are applied recursively. If ``True``, all residual
                blocks at a given level share their convolutional modules. Other modules such as normalization
                layers are not shared. When ``recursive=False``, a new residual block without any shared modules
                is created each time. When ``recursive=True`` and the number of feature channels or spatial size
                of a given residual block does not match the number of output channels or spatial size of the
                preceeding block, respectively, a pre-convolutional layer which adjusts the number of channels
                and/or spatial size is inserted between these residual blocks even when ``pre_conv=False``.
            pre_conv: Always insert a separate convolutional layer between levels. When ``recursive=True``,
                this is also the case when ``pre_conv=False`` if the output and input tensor shapes of final
                and first residual block in subsequent levels do not match.
            post_deconv: Whether to place upsampling layers after the sequence of residual blocks for levels
                with a ``stride`` that is less than 1. By default, upsampling is performed as part of the
                ``pre_conv``. If ``post_deconv=False``, a pre-upsampling layer is always inserted when a
                level has an initial stride of less than 1 regardless of the ``pre_conv`` setting.
            bias: Whether to use bias terms of convolutional layers. Can be either a boolean, or a string
                indicating the function used to initialize these bias terms (cf. ``ConvLayer``).
            norm: Type of normalization to use in convolutional layers. Use no normalization if ``None``.
            acti: Type of non-linear activation function to use in each convolutional layer.
            order: Order of convolution (C), normalization (N), and non-linear activation (A) in each
                convolutional layer. If this string does not contain the character ``n|N``, no normalization
                is performed regardless of the setting of ``norm`` (cf. ``ConvLayer``).
            skip: Type(s) of shortcut connections (cf. ``ResidualUnit``).
            residual_pre_conv: Type of pre-convolution when residual unit is a bottleneck block.
                The kernel size is set to 1 if "conv1", and ``kernel_size`` otherwise.
            residual_post_conv: Type of post-convolution when residual unit is a bottleneck block.
            conv_layer: Type or callable used to create convolutional layers (cf. ``pre_conv``).
            deconv_layer: Type or callable used to create upsampling layers.
            resnet_block: Type or callable used to create residual blocks.
            input_layer: Type or callable used to create initial layer which receives the input tensor.
            output_layer: Type or callable used to create an output layer (head). If ``None`` and
                ``num_classes`` is specified, a default ``classification_head`` is added.

        """
        super().__init__()
        padding_mode = PaddingMode(padding_mode)
        order = order.upper()
        if "C" not in order:
            raise ValueError("ResNet() 'order' must contain 'C' for convolution")
        if isinstance(acti, str):
            nonlinearity = acti
        elif isinstance(acti, Sequence) and len(acti) == 2:
            nonlinearity = acti[0]
            if not isinstance(nonlinearity, str):
                raise TypeError("ResNet() 'acti[0]' must be str")
        else:
            raise ValueError("ResNet() 'acti' must be str or 2-tuple")
        self.nonlinearity = nonlinearity  # used by self.init_conv_modules()
        # Number of residual blocks determined by longest sequence (channels, blocks, stride)
        if isinstance(num_channels, int):
            num_channels = (num_channels,)
        if isinstance(num_blocks, int):
            num_blocks = (num_blocks,)
        if isinstance(stride, (float, int)):
            stride = (stride,)
        # Initial layer (optional)
        if input_layer is None:
            channels = in_channels
        else:
            channels = max(1, int(num_channels[0] * expansion + 0.5))
            layer = input_layer(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=channels,
                kernel_size=kernel_size,
                padding_mode=padding_mode,
                bias=bias,
                norm=norm,
                acti=acti,
                order=order,
            )
            self.add_module("input_layer", layer)
        # Cascade of residual blocks
        for i, (m, b, s) in enumerate(zip_longest_repeat_last(num_channels, num_blocks, stride)):
            if m < 1:
                raise ValueError(f"ResNet() 'num_channels' must be positive, got {m}")
            n = max(1, int(m * expansion + 0.5))
            deconv = None
            if s < 1:
                deconv_in_channels = n if post_deconv else channels
                deconv_out_channels = deconv_in_channels
                if pre_conv and not post_deconv:
                    deconv_out_channels = n
                    channels = n
                deconv = deconv_layer(
                    spatial_dims=spatial_dims,
                    in_channels=deconv_in_channels,
                    out_channels=deconv_out_channels,
                    scale_factor=1 / s,
                    bias=bias,
                )
                if not post_deconv:
                    self.add_module(f"deconv_{i}", deconv)
                s = 1
            with_pre_conv = pre_conv and (i > 0 or input_layer is None)
            with_pre_conv = with_pre_conv or (recursive and (s != 1 or channels != n))
            with_pre_conv = with_pre_conv and (deconv is None or post_deconv)
            if with_pre_conv:
                is_first = input_layer is None and i == 0
                conv = conv_layer(
                    level=i,
                    is_first=is_first,
                    spatial_dims=spatial_dims,
                    in_channels=channels,
                    out_channels=n,
                    kernel_size=kernel_size,
                    padding_mode=padding_mode,
                    stride=s,
                    bias=bias,
                    norm=norm,
                    acti=acti,
                    order=order,
                )
                self.add_module(f"conv_{i}", conv)
                channels = n
                s = 1
            block = None
            for j in range(b):
                block = resnet_block(
                    spatial_dims=spatial_dims,
                    in_channels=channels,
                    out_channels=n,
                    num_channels=m,
                    num_layers=num_layers,
                    kernel_size=kernel_size,
                    pre_conv=residual_pre_conv,
                    post_conv=residual_post_conv,
                    padding_mode=padding_mode,
                    stride=s,
                    bias=bias,
                    norm=norm,
                    acti=acti,
                    order=order,
                    skip=skip,
                    other=block if recursive else None,
                )
                channels = n
                s = 1
                self.add_module(f"block_{i}_{j}", block)
            if deconv is not None and post_deconv:
                self.add_module(f"deconv_{i}", deconv)
        # Optional output layer (e.g., classification head)
        if num_classes and output_layer is None:
            output_layer = classification_head
        if output_layer is not None:
            head_kwargs = dict(
                kernel_size=kernel_size,
                padding_mode=padding_mode,
                bias=bias,
                norm=norm,
                acti=acti,
                order=order,
            )
            if num_classes:
                head_kwargs["num_classes"] = num_classes
            head = output_layer(spatial_dims, channels, **head_kwargs)
            self.add_module("output_layer", head)
        self.spatial_dims = spatial_dims
        self.in_channels = in_channels
        self.out_channels = channels
        self.num_classes = num_classes or None

    @classmethod
    def from_config(cls: Type[ResNet], config: ResNetConfig) -> ResNet:
        return cls(
            spatial_dims=config.spatial_dims,
            in_channels=config.in_channels,
            num_channels=config.num_channels,
            num_blocks=config.num_blocks,
            num_layers=config.num_layers,
            num_classes=config.num_classes,
            kernel_size=config.kernel_size,
            stride=config.stride,
            expansion=config.expansion,
            padding_mode=config.padding_mode,
            recursive=config.recursive,
            pre_conv=config.pre_conv,
            post_deconv=config.post_deconv,
            bias=config.bias,
            norm=config.norm,
            acti=config.acti,
            order=config.order,
            skip=config.skip,
            residual_pre_conv=config.residual_pre_conv,
            residual_post_conv=config.residual_post_conv,
        )

    @classmethod
    def from_dict(cls: Type[ResNet], config: Mapping[str, Any]) -> ResNet:
        config = ResNetConfig.from_dict(config)
        return cls.from_config(config)

    @classmethod
    def from_depth(
        cls: Type[ResNet],
        model_depth: int,
        spatial_dims: int,
        in_channels: int = 1,
        **kwargs,
    ) -> ResNet:
        config = ResNetConfig.from_depth(model_depth, spatial_dims, in_channels, **kwargs)
        return cls.from_config(config)

    def init_conv_modules(self) -> ResNet:
        r"""Initialize parameters of convolutions."""
        init_conv_modules(self, nonlinearity=self.nonlinearity)
        return self

    def init_norm_layers(self) -> ResNet:
        r"""Initialize normalization layer weights and biases."""
        init_norm_layers(self)
        return self

    def zero_init_residuals(self) -> ResNet:
        r"""Zero-initialize the last normalization layer in each residual branch."""
        zero_init_residuals(self)
        return self


def init_conv_modules(network: Module, nonlinearity: str = "relu") -> Module:
    r"""Initialize parameters of convolutions."""
    # https://github.com/pytorch/vision/blob/5905de086dcebff606f9110d15ebd3d1de18fc2e/torchvision/models/resnet.py#L189-L190
    nonlinearity = nonlinearity.lower()
    if nonlinearity == "lrelu":
        nonlinearity = "leaky_relu"
    if nonlinearity in ("relu", "leaky_relu"):
        for module in network.modules():
            if is_convolution(module):
                init.kaiming_normal_(module.weight, mode="fan_out", nonlinearity=nonlinearity)
                if module.bias is not None:
                    init.constant_(module.bias, 0)
    return network


def init_norm_layers(network: Module) -> Module:
    r"""Initialize batch and group norm layers weights to one and biases to zero."""
    # https://github.com/pytorch/vision/blob/5905de086dcebff606f9110d15ebd3d1de18fc2e/torchvision/models/resnet.py#L191-L193
    for module in network.modules():
        if is_batch_norm(module) or is_group_norm(module):
            init.constant_(module.weight, 1)
            init.constant_(module.bias, 0)
    return network


def zero_init_residuals(network: Module) -> Module:
    r"""Zero-initialize the last normalization layer in each residual branch."""
    # https://github.com/pytorch/vision/blob/5905de086dcebff606f9110d15ebd3d1de18fc2e/torchvision/models/resnet.py#L195-L203
    for module in network.modules():
        if isinstance(module, ResidualUnit):
            module.zero_init_residual()
    return network



r"""Optimization-based pairwise image registration."""

from pathlib import Path
from timeit import default_timer as timer
from typing import Any, Dict, Mapping, Optional, Set, Sequence, Tuple, Union

import torch
from torch import Tensor
from torch.nn import Module

from deepali.core import Axes, Device, Grid, PathStr
from deepali.core import functional as U
from deepali.core import join_kwargs_in_sequence
from deepali.data import FlowField, Image
from deepali.losses import RegistrationResult, new_loss
from deepali.spatial import DisplacementFieldTransform
from deepali.spatial import HomogeneousTransform
from deepali.spatial import NonRigidTransform
from deepali.spatial import QuaternionRotation
from deepali.spatial import RigidQuaternionTransform
from deepali.spatial import SequentialTransform
from deepali.spatial import SpatialTransform
from deepali.spatial import Translation
from deepali.spatial import new_spatial_transform

from .engine import RegistrationEngine
from .hooks import RegistrationEvalHook, RegistrationStepHook
from .hooks import normalize_grad_hook, smooth_grad_hook
from .losses import PairwiseImageRegistrationLoss, weight_channel_names
from .optim import new_optimizer


def register_pairwise(
    target: Union[PathStr, Dict[str, PathStr]],
    source: Union[PathStr, Dict[str, PathStr]],
    config: Optional[Dict[str, Any]] = None,
    outdir: Optional[PathStr] = None,
    verbose: Union[bool, int] = False,
    debug: Union[bool, int] = False,
    device: Optional[Device] = None,
) -> SpatialTransform:
    r"""Register pair of images using ``torch.autograd`` and ``torch.optim``."""
    # Configuration
    if config is None:
        config = {}
    if outdir is not None:
        outdir = Path(outdir).absolute()
        outdir.mkdir(parents=True, exist_ok=True)
    loss_config, loss_weights = get_loss_config(config)
    model_name, model_args, model_init = get_model_config(config)
    optim_name, optim_args, optim_loop = get_optim_config(config)
    levels, coarsest_level, finest_level = get_levels_config(config)
    finest_spacing, min_size, pyramid_dims = get_pyramid_config(config)
    device = get_device_config(config, device)
    verbose = int(verbose)
    debug = int(debug)
    if verbose > 0:
        print()
    # Read input images
    start = timer()
    target_keys = set(loss_config.keys()) | set(weight_channel_names(loss_weights).values())
    target_image, target_chns = read_images(target, names=target_keys, device=device)
    source_image, source_chns = read_images(source, names=loss_config.keys(), device=device)
    if verbose > 3:
        print(f"Read images from files in {timer() - start:.3f}s")
    start_reg = timer()
    # Append foreground masks
    target_image = append_mask(target_image, target_chns, config)
    source_image = append_mask(source_image, source_chns, config)
    # Clamp and rescale images
    norm_params = get_normalize_config(config, target_image, target_chns)
    target_image = normalize_data_(target_image, target_chns, **norm_params)
    source_image = normalize_data_(source_image, source_chns, **norm_params)
    # Create Gaussian image resolution pyramids
    start = timer()
    target_pyramid = target_image.pyramid(
        levels,
        start=finest_level,
        end=coarsest_level,
        dims=pyramid_dims,
        spacing=finest_spacing,
        min_size=min_size,
    )
    source_pyramid = source_image.pyramid(
        levels,
        start=finest_level,
        end=coarsest_level,
        dims=pyramid_dims,
        spacing=finest_spacing,
        min_size=min_size,
    )
    if verbose > 3:
        print(f"Constructed Gaussian resolution pyramids in {timer() - start:.3f}s\n")
    if verbose > 2:
        print("Target image pyramid:")
        print_pyramid_info(target_pyramid)
        print("Source image pyramid:")
        print_pyramid_info(source_pyramid)
    # Free no longer needed images
    del target_image
    del source_image
    # Initialize transformation
    source_grid = source_pyramid[finest_level].grid()
    finest_grid = target_pyramid[finest_level].grid()
    coarsest_grid = target_pyramid[coarsest_level].grid()
    post_transform = get_post_transform(config, finest_grid, source_grid)
    transform_downsample = model_args.pop("downsample", 0)
    transform_grid = coarsest_grid.downsample(transform_downsample)
    transform = new_spatial_transform(model_name, grid=transform_grid, groups=1, **model_args)
    if model_init:
        if verbose > 1:
            print(f"Fitting '{model_init}'...")
        disp_field = FlowField.read(model_init).to(device=device)
        assert isinstance(disp_field, FlowField)
        start = timer()
        transform = transform.to(device=device).fit(disp_field.batch())
        if verbose > 0:
            print(f"Fitted initial displacement field in {timer() - start:.3f}s")
        del disp_field
    grid_transform = SequentialTransform(transform, post_transform)
    grid_transform = grid_transform.to(device=device)
    # Perform coarse-to-fine multi-resolution registration
    for level in range(coarsest_level, finest_level - 1, -1):
        target_image = target_pyramid[level]
        source_image = source_pyramid[level]
        # Write image data for debugging
        if outdir and debug > 0:
            write_channels(
                data=target_image.tensor(),
                grid=target_image.grid(),
                channels=target_chns,
                outdir=outdir,
                prefix=f"level_{level}_target_",
            )
            write_channels(
                data=source_image.tensor(),
                grid=source_image.grid(),
                channels=source_chns,
                outdir=outdir,
                prefix=f"level_{level}_source_",
            )
        # Initialize transformation
        if level != coarsest_level:
            start = timer()
            transform_grid = target_image.grid().downsample(transform_downsample)
            transform.grid_(transform_grid)
            if verbose > 3:
                print(f"Subdivided control point grid in {timer() - start:.3f}s")
        grid_transform.grid_(target_image.grid())
        # Registration loss function
        loss_terms = new_loss_terms(loss_config)
        loss = PairwiseImageRegistrationLoss(
            losses=loss_terms,
            source_data=source_image.tensor().unsqueeze(0),
            target_data=target_image.tensor().unsqueeze(0),
            source_grid=source_image.grid(),
            target_grid=target_image.grid(),
            source_chns=source_chns,
            target_chns=target_chns,
            transform=grid_transform,
            weights=loss_weights,
        )
        loss = loss.to(device=device)
        # Write initial source image and flow field
        if outdir and debug > 1:
            start = timer()
            result = loss.eval()
            if verbose > 3:
                print(f"Evaluated initial loss in {timer() - start:.3f}s")
            write_result(
                result,
                grid=target_image.grid(),
                channels=source_chns,
                outdir=outdir,
                prefix=f"level_{level}_initial_",
            )
            flow = grid_transform.flow(target_image.grid(), device=device)
            flow[0].write(outdir / f"level_{level}_initial_def.mha")
        # Initialize optimizer
        optimizer = new_optimizer(optim_name, model=grid_transform, **optim_args)
        # Perform registration at current resolution level
        engine = RegistrationEngine(
            loss=loss,
            optimizer=optimizer,
            max_steps=optim_loop.get("max_steps", 250),
            min_delta=float(optim_loop.get("min_delta", "nan")),
        )
        grad_sigma = float(optim_loop.get("smooth_grad", 0))
        if isinstance(transform, NonRigidTransform) and grad_sigma > 0:
            engine.register_eval_hook(smooth_grad_hook(transform, sigma=grad_sigma))
        engine.register_eval_hook(normalize_grad_hook(transform))
        if verbose > 2:
            engine.register_eval_hook(print_eval_loss_hook(level))
        elif verbose > 1:
            engine.register_step_hook(print_step_loss_hook(level))
        if outdir and debug > 2:
            engine.register_eval_hook(
                write_result_hook(
                    level=level, grid=target_image.grid(), channels=source_chns, outdir=outdir
                )
            )
        engine.run()
        # Report result
        if verbose > 0 or (outdir and debug > 0):
            start = timer()
            result = loss.eval()
            if verbose > 3:
                print(f"Evaluated final loss in {timer() - start:.3f}s")
            if verbose > 0:
                loss_value = float(result["loss"])
                print(
                    f"level={level:d}: loss={loss_value:.5f} ({engine.num_steps:d} steps)",
                    flush=True,
                )
            # Resulting transform of current level
            if outdir and debug > 0:
                write_result(
                    result,
                    grid=target_image.grid(),
                    channels=source_chns,
                    outdir=outdir,
                    prefix=f"level_{level}_final_",
                )
                flow = grid_transform.flow(device=device)
                flow[0].write(outdir / f"level_{level}_final_def.mha")
    if verbose > 3:
        print(f"Registered images in {timer() - start_reg:.3f}s")
    if verbose > 0:
        print()
    return grid_transform


def append_mask(
    image: Image,
    channels: Dict[str, Tuple[int, int]],
    config: Dict[str, Any],
) -> Image:
    r"""Append foreground mask to data tensor."""
    data = image.tensor()
    if "img" in channels:
        lower_threshold, upper_threshold = get_clamp_config(config, "img")
        mask = U.threshold(data[slice(*channels["img"])], lower_threshold, upper_threshold)
    else:
        mask = torch.ones((1,) + data.shape[1:], dtype=data.dtype, device=data.device)
    data = torch.cat([data, mask.type(data.dtype)], dim=0)
    channels["msk"] = (data.shape[0] - 1, data.shape[0])
    return Image(data, image.grid())


def append_data(
    data: Optional[Tensor], channels: Dict[str, Tuple[int, int]], name: str, other: Tensor
) -> Tensor:
    r"""Append image data."""
    if data is None:
        data = other
    else:
        data = torch.cat([data, other], dim=0)
    channels[name] = (data.shape[0] - other.shape[0], data.shape[0])
    return data


def read_images(
    sample: Union[PathStr, Dict[str, PathStr]], names: Set[str], device: torch.device
) -> Tuple[Image, Dict[str, Tuple[int, int]]]:
    r"""Read image data from input files."""
    data = None
    grid = None
    if isinstance(sample, (Path, str)):
        sample = {"img": sample}
    img_path = sample.get("img")
    seg_path = sample.get("seg")
    sdf_path = sample.get("sdf")
    for path in (img_path, seg_path, sdf_path):
        if not path:
            continue
        grid = Grid.from_file(path).align_corners_(True)
        break
    else:
        raise ValueError("One of 'img', 'seg', or 'sdf' input image file paths is required")
    assert grid is not None
    dtype = torch.float32
    channels = {}
    if "img" in names:
        temp = Image.read(img_path, dtype=dtype, device=device)
        data = append_data(data, channels, "img", temp.tensor())
    if "seg" in names:
        if seg_path is None:
            raise ValueError("Missing segmentation label image file path")
        temp = Image.read(seg_path, dtype=torch.int64, device=device)
        temp_grid = temp.grid()
        num_classes = int(temp.max()) + 1
        temp = temp.tensor().unsqueeze(0)
        temp = U.as_one_hot_tensor(temp, num_classes).to(dtype=dtype)
        temp = temp.squeeze(0)
        temp = Image(temp, grid=temp_grid).sample(grid)
        data = append_data(data, channels, "seg", temp.tensor())
    if "sdf" in names:
        if sdf_path is None:
            raise ValueError("Missing segmentation boundary signed distance field file path")
        temp = Image.read(sdf_path, dtype=dtype, device=device)
        temp = temp.sample(grid)
        data = append_data(data, channels, "sdf", temp.tensor())
    if data is None:
        if img_path is None:
            raise ValueError("Missing intensity image file path")
        data = Image.read(img_path, dtype=dtype, device=device)
        channels = {"img": (0, 1)}
    image = Image(data, grid=grid)
    return image, channels


def get_device_config(
    config: Dict[str, Any], device: Optional[Union[str, torch.device]] = None
) -> torch.device:
    r"""Get configured PyTorch device."""
    if device is None:
        device = config.get("device", "cpu")
    if isinstance(device, int):
        device = f"cuda:{device}"
    elif device == "cuda":
        device = "cuda:0"
    return torch.device(device)


def load_transform(path: PathStr, grid: Grid) -> SpatialTransform:
    r"""Load transformation from file.

    Args:
        path: File path from which to load spatial transformation.
        grid: Target domain grid with respect to which transformation is defined.

    Returns:
        Loaded spatial transformation.

    """
    target_grid = grid

    def convert_matrix(matrix: Tensor, grid: Optional[Grid] = None) -> Tensor:
        if grid is None:
            pre = target_grid.transform(Axes.CUBE_CORNERS, Axes.WORLD)
            post = target_grid.transform(Axes.WORLD, Axes.CUBE_CORNERS)
            matrix = U.homogeneous_matmul(post, matrix, pre)
        elif grid != target_grid:
            pre = target_grid.transform(Axes.CUBE_CORNERS, grid=grid)
            post = grid.transform(Axes.CUBE_CORNERS, grid=target_grid)
            matrix = U.homogeneous_matmul(post, matrix, pre)
        return matrix

    path = Path(path)
    if path.suffix == ".pt":
        value = torch.load(path, map_location="cpu")
        if isinstance(value, dict):
            matrix = value.get("matrix")
            if matrix is None:
                raise KeyError("load_transform() .pt file dict must contain key 'matrix'")
            grid = value.get("grid")
        elif isinstance(value, Tensor):
            matrix = value
            grid = None
        else:
            raise RuntimeError("load_transform() .pt file must contain tensor or dict")
        if matrix.ndim == 2:
            matrix = matrix.unsqueeze(0)
        if matrix.ndim != 3 or matrix.shape[1:] != (3, 4):
            raise RuntimeError("load_transform() .pt file tensor must have shape (N, 3, 4)")
        params = convert_matrix(matrix, grid)
        return HomogeneousTransform(target_grid, params=params)

    flow = FlowField.read(path, axes=Axes.WORLD)
    flow = flow.axes(Axes.from_grid(target_grid))
    flow = flow.sample(target_grid)
    return DisplacementFieldTransform(target_grid, params=flow.tensor().unsqueeze(0))


def get_post_transform(
    config: Dict[str, Any],
    target_grid: Grid,
    source_grid: Grid,
) -> Optional[SpatialTransform]:
    r"""Get constant rigid transformation between image grid domains."""
    align = config.get("align", False)
    if align is False or align is None:
        return None
    if isinstance(align, (Path, str)):
        return load_transform(align, target_grid)
    if align is True:
        align_centers = True
        align_directions = True
    elif isinstance(align, dict):
        align_centers = bool(align.get("centers", True))
        align_directions = bool(align.get("directions", True))
    else:
        raise ValueError("get_post_transform() 'config' has invalid 'align' value: {align}")
    center_offset = (
        target_grid.world_to_cube(source_grid.center()).unsqueeze(0) if align_centers else None
    )
    rotation_matrix = (
        source_grid.direction() @ target_grid.direction().t().unsqueeze(0)
        if align_directions
        else None
    )
    transform = None
    if center_offset is not None and rotation_matrix is not None:
        transform = RigidQuaternionTransform(target_grid, translation=center_offset, rotation=False)
        transform.rotation.matrix_(rotation_matrix)
    elif center_offset is not None:
        transform = Translation(target_grid, params=center_offset)
    elif rotation_matrix is not None:
        transform = QuaternionRotation(target_grid, params=False)
        transform.matrix_(rotation_matrix)
    return transform


def get_clamp_config(
    config: Dict[str, Any], channel: str
) -> Tuple[Optional[float], Optional[float]]:
    r"""Get thresholds for named image channel.

    Args:
        config: Configuration.
        channel: Name of image channel.

    Returns:
        lower_threshold: Lower threshold.
        upper_threshold: Upper threshold.

    """
    input_config = config.get("input", {})
    if not isinstance(input_config, dict):
        raise ValueError("get_clamp_config() 'input' value must be dict")
    channel_config = input_config.get(channel)
    if not isinstance(channel_config, dict):
        channel_config = {"clamp": channel_config}
    thresholds = channel_config.get("clamp", input_config.get("clamp"))
    if thresholds is None:
        thresholds = (None, None)
    elif isinstance(thresholds, (int, float)):
        thresholds = (float(thresholds), None)
    if not isinstance(thresholds, (list, tuple)):
        raise ValueError("get_clamp_config() value must be scalar or sequence")
    if len(thresholds) != 2:
        raise ValueError("get_clamp_config() value must be scalar or [min, max]")
    thresholds = tuple((None if v is None else float(v) for v in thresholds))
    lower_threshold, upper_threshold = thresholds
    return lower_threshold, upper_threshold


def get_scale_config(config: Dict[str, Any], channel: str) -> Optional[float]:
    r"""Get channel scaling factor."""
    input_config = config.get("input", {})
    if not isinstance(input_config, dict):
        return None
    channel_config = input_config.get(channel)
    if not isinstance(channel_config, dict):
        return None
    value = channel_config.get("scale", input_config.get("scale"))
    if value is None:
        return None
    return float(value)


def get_normalize_config(
    config: Dict[str, Any], image: Image, channels: Dict[str, Tuple[int, int]]
) -> Dict[str, Dict[str, Tensor]]:
    r"""Calculate data normalization parameters.

    Args:
        config: Configuration.
        image: Image data.
        channels: Map of image channel slices.

    Returns:
        Dictionary of normalization parameters.

    """
    scale = {}
    shift = {}
    for channel, (start, stop) in channels.items():
        data = image.tensor().narrow(0, start, stop - start)
        lower_threshold, upper_threshold = get_clamp_config(config, channel)
        scale_factor = get_scale_config(config, channel)
        if channel in ("msk", "seg"):
            if lower_threshold is None:
                lower_threshold = 0
            if upper_threshold is None:
                upper_threshold = 1
        else:
            if lower_threshold is None:
                lower_threshold = data.min()
            if upper_threshold is None:
                upper_threshold = data.max()
        if scale_factor is None:
            if upper_threshold > lower_threshold:
                scale_factor = upper_threshold - lower_threshold
            else:
                scale_factor = 1
        else:
            scale_factor = 1 / scale_factor
        shift[channel] = lower_threshold
        scale[channel] = scale_factor
    return dict(shift=shift, scale=scale)


def normalize_data_(
    image: Image,
    channels: Dict[str, Tuple[int, int]],
    shift: Optional[Dict[str, Tensor]] = None,
    scale: Optional[Dict[str, Tensor]] = None,
) -> Image:
    r"""Normalize image data."""
    if shift is None:
        shift = {}
    if scale is None:
        scale = {}
    for channel, (start, stop) in channels.items():
        data = image.tensor().narrow(0, start, stop - start)
        offset = shift.get(channel)
        if offset is not None:
            data -= offset
        norm = scale.get(channel)
        if norm is not None:
            data /= norm
        if channel in ("msk", "seg"):
            data.clamp_(min=0, max=1)
    return image


def get_levels_config(config: Dict[str, Any]) -> Tuple[int, int, int]:
    r"""Get indices of coarsest and finest level from configuration."""
    cfg = config.get("pyramid", {})
    levels = cfg.get("levels", 4)
    if isinstance(levels, int):
        levels = (levels - 1, 0)
    if not isinstance(levels, (list, tuple)):
        raise TypeError(
            "register_pairwise() 'config' key 'pyramid.levels': value must be int, tuple, or list"
        )
    coarsest_level, finest_level = levels
    if finest_level > coarsest_level:
        raise ValueError(
            "register_pairwise() 'config' key 'pyramid.levels':"
            + " finest level must be less or equal than coarsest level"
        )
    levels = coarsest_level + 1
    if "max_level" in cfg:
        levels = max(levels, cfg["max_level"])
    return levels, coarsest_level, finest_level


def get_pyramid_config(
    config: Dict[str, Any]
) -> Tuple[Optional[Union[float, Sequence[float]]], int, Optional[Union[str, int]]]:
    r"""Get settings of Gaussian resolution pyramid from configuration."""
    cfg = config.get("pyramid", {})
    min_size = cfg.get("min_size", 16)
    finest_spacing = cfg.get("spacing")
    dims = cfg.get("dims")
    return finest_spacing, min_size, dims


def get_loss_config(config: Dict[str, Any]) -> Tuple[Dict[str, Dict[str, Any]], Dict[str, float]]:
    r"""Instantiate terms of registration loss given configuration object.

    Args:
        config: Configuration.

    Returns:
        losses: Preparsed configuration of loss terms (cf. ``new_loss_terms()``).
        weights: Weights of loss terms.

    """
    cfg = None
    losses = {}
    weights = {}
    sections = ("loss", "losses", "energy")
    for name in sections:
        if name in config:
            if cfg is not None:
                raise ValueError("get_loss_config() keys {sections} are mutually exclusive")
            cfg = config[name]
    if not cfg:
        cfg = "SSD"
    if isinstance(cfg, str):
        cfg = (cfg,)
    if isinstance(cfg, Sequence):
        names, cfg = cfg, {}
        for i, name in enumerate(names):
            cfg[f"loss_{i}"] = str(name)
    if isinstance(cfg, dict):
        for key, value in cfg.items():
            name = None
            weight = 1
            kwargs = {}
            if isinstance(value, str):
                name = value
            elif isinstance(value, Sequence):
                if not value:
                    raise ValueError(f"get_loss_config() '{key}' loss entry is empty")
                if len(value) == 1:
                    if isinstance(value[0], str):
                        value = {"name": value[0]}
                elif len(value) > 1:
                    if isinstance(value[0], (int, float)):
                        value[0] = {"weight": value[0]}
                    if isinstance(value[1], str):
                        value[1] = {"name": value[1]}
                value = join_kwargs_in_sequence(value)
                if isinstance(value, dict):
                    kwargs = dict(value)
                    name = kwargs.pop("name", None)
                    weight = kwargs.pop("weight", 1)
                elif len(value) == 2:
                    name = value[0]
                    kwargs = dict(value[1])
                elif len(value) == 3:
                    weight = float(value[0])
                    name = value[1]
                    kwargs = dict(value[2])
                else:
                    raise ValueError(f"get_loss_config() '{key}' invalid loss configuration")
            elif isinstance(value, dict):
                kwargs = dict(value)
                name = kwargs.pop("name", None)
                weight = kwargs.pop("weight", 1)
            else:
                weight, name = value
            if name is None:
                raise ValueError(f"get_loss_config() missing 'name' for loss '{key}'")
            if not isinstance(name, str):
                raise TypeError(f"get_loss_config() 'name' of loss '{key}' must be str")
            kwargs["name"] = name
            losses[key] = kwargs
            weights[key] = weight
    else:
        raise TypeError("get_loss_config() 'config' \"losses\" must be str, tuple, list, or dict")
    weights_config = config.get("weights", {})
    if isinstance(weights_config, (int, float)):
        weights_config = (weights_config,)
    if isinstance(weights_config, (list, tuple)):
        names, weights_config = weights_config, {}
        for i, weight in enumerate(names):
            weights_config[f"loss_{i}"] = weight
    if not isinstance(weights_config, dict):
        raise TypeError("get_loss_config() 'weights' must be scalar, tuple, list, or dict")
    weights.update(weights_config)
    losses = {k: v for k, v in losses.items() if weights.get(k, 0)}
    weights = {k: v for k, v in weights.items() if k in losses}
    return losses, weights


def new_loss_terms(config: Dict[str, Any]) -> Dict[str, Module]:
    r"""Instantiate terms of registration loss.

    Args:
        config: Preparsed configuration of loss terms.
        target_tree: Target vessel centerline tree.

    Returns:
        Mapping from channel or loss name to loss module instance.

    """
    losses = {}
    for key, value in config.items():
        kwargs = dict(value)
        name = kwargs.pop("name", None)
        _ = kwargs.pop("weight", None)
        if name is None:
            raise ValueError(f"new_loss_terms() missing 'name' for loss '{key}'")
        if not isinstance(name, str):
            raise TypeError(f"new_loss_terms() 'name' of loss '{key}' must be str")
        loss = new_loss(name, **kwargs)
        losses[key] = loss
    return losses


def get_model_config(config: Dict[str, Any]) -> Tuple[str, Dict[str, Any], Optional[str]]:
    r"""Get configuration of transformation model to use."""
    cfg = config.get("model", {})
    cfg = dict(name=cfg) if isinstance(cfg, str) else dict(cfg)
    # Name of transformation
    model_name = cfg.pop("name")
    assert isinstance(model_name, str)
    assert model_name != ""
    # Initial guess
    model_init = cfg.pop("init", None)
    if model_init is not None:
        config_path = config["__file__"]
        assert isinstance(config_path, Path)
        model_init = (config_path.parent / model_init).as_posix()
    # Transformation keyword arguments
    model_args = dict(cfg.get(model_name, cfg))
    # optim_args.update(cfg)
    return model_name, model_args, model_init


def get_optim_config(config: Dict[str, Any]) -> Tuple[str, Dict[str, Any], Dict[str, Any]]:
    r"""Get configuration of optimizer to use."""
    cfg = config.get("optim", {})
    cfg = dict(name=cfg) if isinstance(cfg, str) else dict(cfg)
    # Name of optimizer
    if "optimizer" in cfg:
        if "name" in cfg:
            raise ValueError("get_optim_config() keys ('name', 'optimizer') are mutually exclusive")
        cfg["name"] = cfg.pop("optimizer")
    optim_name = str(cfg.pop("name", "LBFGS"))
    # Optimization loop parameters
    optim_loop = {}
    for key in ("max_steps", "min_delta", "smooth_grad"):
        if key in cfg:
            optim_loop[key] = cfg.pop(key)
    # Optimizer keyword arguments (excl. optimizer class name keys starting with uppercase letter)
    optim_args = {k: v for k, v in cfg.items() if isinstance(k, str) and k[0].islower()}
    optim_args.update(cfg.get(optim_name, {}))
    lr_keys = ("step_size", "learning_rate")
    for lr_key in ("step_size", "learning_rate"):
        if lr_key in optim_args:
            if "lr" in optim_args:
                raise ValueError(
                    f"get_optim_config() keys {lr_keys + ('lr',)} are mutually exclusive"
                )
            optim_args["lr"] = optim_args.pop(lr_key)
    return optim_name, optim_args, optim_loop


@torch.no_grad()
def write_channels(
    data: Tensor,
    grid: Grid,
    channels: Mapping[str, Tuple[int, int]],
    outdir: PathStr,
    prefix: str = "",
) -> None:
    r"""Write image channels."""
    for name, (start, stop) in channels.items():
        image = data[slice(start, stop, 1)]
        if name == "seg":
            image = image.argmax(dim=0, keepdim=True).type(torch.uint8)
        elif name == "msk":
            image = image.mul(255).clamp_(0, 255).type(torch.uint8)
        if not isinstance(image, Image):
            image = Image(image, grid=grid)
        image.write(outdir / f"{prefix}{name}.mha")


@torch.no_grad()
def write_result(
    result: RegistrationResult,
    grid: Grid,
    channels: Mapping[str, Tuple[int, int]],
    outdir: PathStr,
    prefix: str = "",
) -> None:
    r"""Write registration result to output directory."""
    # Warped images
    data = result["source"]
    assert isinstance(data, Tensor)
    write_channels(data[0], grid=grid, channels=channels, outdir=outdir, prefix=prefix)
    # Overlap mask
    data = result["mask"]
    assert isinstance(data, Tensor)
    if data.dtype == torch.bool:
        data = data.type(torch.uint8).mul_(255)
    mask = Image(data[0], grid=grid)
    mask.write(outdir / f"{prefix}olm.mha")


def write_result_hook(
    level: int, grid: Grid, channels: Mapping[str, Tuple[int, int]], outdir: Path
) -> RegistrationEvalHook:
    r"""Get callback function for writing results after each evaluation."""

    def fn(
        _: RegistrationEngine,
        num_steps: int,
        num_evals: int,
        result: RegistrationResult,
    ) -> None:
        prefix = f"level_{level}_step_{num_steps:03d}_eval_{num_evals}_"
        write_result(result, grid=grid, channels=channels, outdir=outdir, prefix=prefix)

    return fn


def print_eval_loss_hook(level: int) -> RegistrationEvalHook:
    r"""Get callback function for printing loss after each evaluation."""

    def fn(
        _: RegistrationEngine, num_steps: int, num_eval: int, result: RegistrationResult
    ) -> None:
        loss = float(result["loss"])
        message = f"  {num_steps:>4d}:"
        message += f" {loss:>12.05f} (loss)"
        weights: Dict[str, Union[str, float]] = result.get("weights", {})
        losses: Dict[str, Tensor] = result["losses"]
        for name, value in losses.items():
            value = float(value)
            weight = weights.get(name, 1.0)
            if not isinstance(weight, str):
                value *= weight
            elif "+" in weight:
                weight = f"({weight})"
            message += f", {value:>12.05f} [{weight} * {name}]"
        if num_eval > 1:
            message += " [evals={num_eval:d}]"
        print(message, flush=True)

    return fn


def print_step_loss_hook(level: int) -> RegistrationStepHook:
    r"""Get callback function for printing loss after each step."""

    def fn(_: RegistrationEngine, num_steps: int, num_eval: int, loss: float) -> None:
        message = f"  {num_steps:>4d}: {loss:>12.05f}"
        if num_eval > 1:
            message += " [evals={num_eval:d}]"
        print(message, flush=True)

    return fn


def print_pyramid_info(pyramid: Dict[str, Image]) -> None:
    r"""Print information of image resolution pyramid."""
    levels = sorted(pyramid.keys())
    for level in reversed(levels):
        grid = pyramid[level].grid()
        size = ", ".join([f"{n:>3d}" for n in grid.size()])
        origin = ", ".join([f"{n:.2f}" for n in grid.origin()])
        extent = ", ".join([f"{n:.2f}" for n in grid.extent()])
        domain = ", ".join([f"{n:.2f}" for n in grid.cube_extent()])
        print(
            f"- Level {level}:"
            + f" size=({size})"
            + f", origin=({origin})"
            + f", extent=({extent})"
            + f", domain=({domain})"
        )
    print()



r"""Functions for collating dataset samples containing tensor decorators."""

from collections import abc
from dataclasses import is_dataclass
from pathlib import Path
from typing import Any, List, Mapping, NamedTuple, Sequence, overload

import torch
from torch.utils.data.dataloader import default_collate

from deepali.core.typing import Batch, Dataclass, Sample, is_namedtuple

from .flow import FlowField, FlowFields
from .image import Image, ImageBatch
from .sample import sample_field_names, sample_field_value
from .sample import replace_all_sample_field_values


__all__ = ("collate_samples",)


@overload
def collate_samples(batch: Sequence[Mapping[str, Any]]) -> Mapping[str, Any]:
    ...


@overload
def collate_samples(batch: Sequence[Dataclass]) -> Dataclass:
    ...


@overload
def collate_samples(batch: Sequence[NamedTuple]) -> NamedTuple:
    ...


def collate_samples(batch: Sequence[Sample]) -> Batch:
    r"""Collate individual samples into a batch."""
    if not batch:
        raise ValueError("collate_samples() 'batch' must have at least one element")
    item0 = batch[0]
    names = sample_field_names(item0)
    values = []
    for name in names:
        elem0 = sample_field_value(item0, name)
        samples = [elem0]
        is_none = elem0 is None
        for item in batch[1:]:
            value = sample_field_value(item, name)
            if (is_none and value is not None) or (not is_none and value is None):
                raise ValueError(
                    f"collate_samples() 'batch' has some samples with '{name}' set and others without"
                )
            if not isinstance(value, type(elem0)):
                raise TypeError(
                    f"collate_samples() all 'batch' samples for field '{name}' must be of the same type"
                )
            samples.append(value)
        if is_none:
            values.append(None)
        elif isinstance(elem0, FlowField):
            FlowFieldsType = type(elem0.batch())
            flow_fields: List[FlowField] = samples
            if any(flow_field.axes() != elem0.axes() for flow_field in flow_fields):
                raise ValueError(
                    f"collate_samples() 'batch' contains '{name}' flow fields with mixed axes"
                )
            data = default_collate([flow_field.tensor() for flow_field in flow_fields])
            grid = tuple(flow_field.grid() for flow_field in flow_fields)
            values.append(FlowFieldsType(data, grid, elem0.axes()))
        elif isinstance(elem0, FlowFields):
            FlowFieldsType = type(elem0)
            flow_fields: List[FlowFields] = samples
            if any(flow_field.axes() != elem0.axes() for flow_field in flow_fields):
                raise ValueError(
                    f"collate_samples() 'batch' contains '{name}' flow fields with mixed axes"
                )
            data = torch.cat([flow_field.tensor() for flow_field in flow_fields], dim=0)
            grid = tuple(grid for flow_field in flow_fields for grid in flow_field.grids())
            values.append(FlowFieldsType(data, grid, elem0.axes()))
        elif isinstance(elem0, Image):
            ImageBatchType = type(elem0.batch())
            images: List[Image] = samples
            data = default_collate([image.tensor() for image in images])
            grid = tuple(image.grid() for image in images)
            values.append(ImageBatchType(data, grid))
        elif isinstance(elem0, ImageBatch):
            ImageBatchType = type(elem0)
            images: List[ImageBatch] = samples
            data = torch.cat([image.tensor() for image in images], dim=0)
            grid = tuple(grid for image in images for grid in image.grids())
            values.append(ImageBatchType(data, grid))
        elif isinstance(elem0, (Path, str)):
            values.append(samples)
        elif isinstance(elem0, abc.Mapping) or is_dataclass(elem0) or is_namedtuple(elem0):
            values.append(collate_samples(samples))
        else:
            values.append(default_collate(samples))
    return replace_all_sample_field_values(item0, values)
